"""
Module for interacting with the EdgePi DAC via SPI.
"""


import logging
from edgepi.dac.dac_commands import DACCommands
from edgepi.dac.dac_constants import (
    NULL_BITS,
    NUM_PINS,
    SW_RESET,
    UPPER_LIMIT,
    PowerMode,
    DACChannel,
    EdgePiDacCom as COM,
    EdgePiDacCalibrationConstants as CalibConst,
    AOPins,
    GainPin
)
from edgepi.peripherals.spi import SpiDevice as spi
from edgepi.gpio.edgepi_gpio import EdgePiGPIO
from edgepi.calibration.edgepi_eeprom import EdgePiEEPROM

class EdgePiDAC(spi):
    """A EdgePi DAC device"""

    # analog_out number to pin name
    __analog_out_pin_map = {
        DACChannel.AOUT0.value: AOPins.AO_EN1,
        DACChannel.AOUT1.value: AOPins.AO_EN2,
        DACChannel.AOUT2.value: AOPins.AO_EN3,
        DACChannel.AOUT3.value: AOPins.AO_EN4,
        DACChannel.AOUT4.value: AOPins.AO_EN5,
        DACChannel.AOUT5.value: AOPins.AO_EN6,
        DACChannel.AOUT6.value: AOPins.AO_EN7,
        DACChannel.AOUT7.value: AOPins.AO_EN8,
    }

    def __init__(self):
        self.log = logging.getLogger(__name__)
        self.log.info("Initializing DAC Bus")
        super().__init__(bus_num=6, dev_id=3, mode=1, max_speed=1000000)

        # Read edgepi reserved data and generate calibration parameter dictionary
        eeprom = EdgePiEEPROM()
        eeprom_data  = eeprom.get_edgepi_reserved_data()
        dac_calib_params = eeprom_data.dac_calib_parms

        self.dac_ops = DACCommands(dac_calib_params)
        self.gpio = EdgePiGPIO()

        self.__dac_power_state = {
            DACChannel.AOUT7.value: PowerMode.NORMAL.value,
            DACChannel.AOUT6.value: PowerMode.NORMAL.value,
            DACChannel.AOUT5.value: PowerMode.NORMAL.value,
            DACChannel.AOUT4.value: PowerMode.NORMAL.value,
            DACChannel.AOUT3.value: PowerMode.NORMAL.value,
            DACChannel.AOUT2.value: PowerMode.NORMAL.value,
            DACChannel.AOUT1.value: PowerMode.NORMAL.value,
            DACChannel.AOUT0.value: PowerMode.NORMAL.value,
        }

    def __send_to_gpio_pins(self, analog_out: int, voltage: float):
        ao_pin = self.__analog_out_pin_map[analog_out].value

        if voltage > 0:
            self.gpio.set_pin_state(ao_pin)
        elif voltage == 0:
            self.gpio.clear_pin_state(ao_pin)
        else:
            raise ValueError("voltage cannot be negative")

    # TODO: Decimal instead of float for precision testing
    def write_voltage(self, analog_out: DACChannel, voltage: float):
        """
        Write a voltage value to an analog out pin. Voltage will be continuously
        transmitted to the analog out pin until a 0 V value is written to it.

        Args:
            `analog_out` (DACChannel): A/D_OUT pin to write a voltage value to.

            `voltage` (float): the voltage value to write, in volts.

        Raises:
            `ValueError`: if voltage has more decimal places than DAC accuracy limit
        """
        self.dac_ops.check_range(analog_out.value, 0, NUM_PINS-1)
        self.dac_ops.check_range(voltage, 0, UPPER_LIMIT)
        dac_gain = CalibConst.DAC_GAIN_FACTOR.value if self.__get_gain_state() else 1
        code = self.dac_ops.voltage_to_code(analog_out.value, voltage, dac_gain)
        self.log.debug(f'Code: {code}')

        if self.gpio.get_pin_direction(self.__analog_out_pin_map[analog_out.value].value):
            self.gpio.clear_pin_state(self.__analog_out_pin_map[analog_out.value].value)
            self.gpio.set_pin_direction_out(self.__analog_out_pin_map[analog_out.value].value)

        # update DAC register
        self.transfer(self.dac_ops.generate_write_and_update_command(analog_out.value, code))
        # send voltage to analog out pin
        self.__send_to_gpio_pins(analog_out.value, voltage)
        return code

    def set_power_mode(self, analog_out: DACChannel, power_mode: PowerMode):
        """
        Set power mode for individual DAC channels to either normal power consumption,
        or low power consumption modes.

        For example, this may be of use when no constant power source available.

        Args:
            `analog_out` (DACChannel): the analog out pin whose power mode will be changed

            `power_mode` (PowerMode): a valid hex code for setting DAC channel power mode
        """
        self.dac_ops.check_range(analog_out.value, 0, NUM_PINS-1)
        self.__dac_power_state[analog_out.value] = power_mode.value
        power_code = self.dac_ops.generate_power_code(self.__dac_power_state.values())
        cmd = self.dac_ops.combine_command(COM.COM_POWER_DOWN_OP.value, NULL_BITS, power_code)
        self.transfer(cmd)

    def reset(self):
        """
        Performs a software reset of the EdgePi DAC to power-on default values,
        and stops all voltage transmissions through pins.
        """
        cmd = self.dac_ops.combine_command(COM.COM_SW_RESET.value, NULL_BITS, SW_RESET)
        self.transfer(cmd)

    def channel_readback(self, analog_out: DACChannel) -> int:
        """
        Readback the input register of DAC.

        Args:
            `analog_out` (DACChannel): the analog out pin to read voltage from
        Return:
            (int): code value stored in the input register, can be used to calculate expected
            voltage
        """
        self.dac_ops.check_range(analog_out.value, 0, NUM_PINS-1)
        # first transfer triggers read mode, second is needed to fetch data
        cmd = self.dac_ops.combine_command(
                COM.COM_READBACK.value, DACChannel(analog_out.value).value, NULL_BITS
            )
        self.transfer(cmd)
        # all zero dummy command to trigger second transfer which
        # contains the DAC register contents.
        read_data = self.transfer([NULL_BITS, NULL_BITS, NULL_BITS])
        self.log.debug(f"reading code {read_data}")
        return self.dac_ops.extract_read_data(read_data)


    def compute_expected_voltage(self, analog_out: DACChannel) -> float:
        """
        Computes expected voltage from the DAC channel corresponding to analog out pin.
        This is not guaranteed to be the voltage of the analog_out pin at the terminal block,
        nor the voltage being output by the DAC channel. It is a calculation that maps
        the contents of the DAC channel to a voltage value.

        Args:
            analog_out (DACChannel): the analog out pin to read voltage from

        Returns:
            float: the computed voltage value of the DAC channel corresponding
                to the selected analog out pin.
        """
        code = self.channel_readback(analog_out)
        dac_gain = CalibConst.DAC_GAIN_FACTOR.value if self.__get_gain_state() else 1
        return self.dac_ops.code_to_voltage(analog_out.value, code, dac_gain)

    def enable_dac_gain(self, enable: bool = None):
        """
        Enable/Disable internal DAC gain by toggling the DAC_GAIN pin
        Args:
            enable (bool): enable boolean to set or clear the gpio pin
        Return:
            gain_state (bool): state of the gain pin
        """
        # pylint: disable=expression-not-assigned
        self.gpio.set_pin_state(GainPin.DAC_GAIN.value) if enable else \
        self.gpio.clear_pin_state(GainPin.DAC_GAIN.value)
        return self.__get_gain_state()

    def __get_gain_state(self):
        """
        Retrieve the internal gain state by reading the expander pin
        Args:
            N/A
        Return:
            gain_state (bool): True - gain enalbed, False - gain disabled
        """
        _, _, gain_state = self.get_state(gain=True)
        return gain_state

    def get_state(self, analog_out: DACChannel = None,
                        code: bool = None,
                        voltage: bool = None,
                        gain: bool = None):
        """
        the method returns the state of requested parameters. It will either read the register of
        DAC or GPIO expander to retrieve the current state.

        Args:
            analog_out (DACChannel): channel number of interest
            code (bool): requesting the current code value written in the specified channel input
                         register
            voltage (bool): requesting the current expected voltage at the terminal block pin
            gian (bool): requesting the current gain value set for the DAC
        Returns:
            code_val (int): code value read from the input register, None when not requested
            voltage_val (float): voltage calculated using the code value, None when not requested
            gain_state (bool): true if dac gain is enabled or False disabled, None when not
                               requested
        """
        code_val = self.channel_readback(analog_out) if code else None
        voltage_val = self.compute_expected_voltage(analog_out) if voltage else None
        gain_state = self.gpio.get_pin_direction(GainPin.DAC_GAIN.value) if gain else None
        return code_val, voltage_val, gain_state



"""unit tests for edgepi_gpio_expander module"""

# pylint: disable=protected-access
# pylint: disable=C0413

from unittest import mock
from unittest.mock import patch
import sys
if sys.platform != 'linux':
    sys.modules['periphery'] = mock.MagicMock()
from copy import deepcopy

import pytest
from edgepi.gpio.gpio_configs import DACPins, generate_expander_pin_info
from edgepi.gpio.edgepi_gpio_expander import EdgePiGPIOExpander

@pytest.fixture(name='mock_i2c')
def fixture_mock_i2c_lib(mocker):
    yield mocker.patch('edgepi.peripherals.i2c.I2C')

@pytest.mark.parametrize("mock_expect, result",
                        [(['/dev/i2c-10'],[]),
                        ])
@patch('edgepi.peripherals.i2c.I2CDevice')
def test_edgepi_expander_init(mock_i2c_device, mock_expect, result, mock_i2c):
    mock_i2c.return_value = None
    mock_i2c_device.fd = mock_expect[0]
    result = generate_expander_pin_info()
    gpio_ctrl = EdgePiGPIOExpander()
    assert gpio_ctrl.expander_pin_dict == result

@pytest.mark.parametrize("dev_address, out",
                        [(32, 255),
                         (33, 255),
                         (33, 255),
                         (32, 255),
                        ])
@patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.set_read_msg')
@patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.transfer')
def test_edgepi_expander_read_register(mock_data, mock_msg, dev_address, out, mock_i2c):
    mock_i2c.return_value = None
    mock_msg.data = [255]
    mock_msg.return_value = (mock_msg ,mock_msg)
    mock_data.return_value = out
    gpio_ctrl = EdgePiGPIOExpander()
    out_data = gpio_ctrl._EdgePiGPIOExpander__read_register(7,
                                                            dev_address)
    assert out_data == out

# TODO: these need to be refactored to work with new methods
@pytest.mark.parametrize("pin_name, mock_value, result",
                         [(DACPins.AO_EN1.value, 170, True),
                          (DACPins.AO_EN2.value, 170, False),
                          (DACPins.AO_EN3.value, 170, True),
                          (DACPins.AO_EN4.value, 170, False),
                          (DACPins.AO_EN5.value, 170, True),
                          (DACPins.AO_EN6.value, 170, False),
                          (DACPins.AO_EN7.value, 170, True),
                          (DACPins.AO_EN8.value, 170, False),
                          (DACPins.AO_EN1.value, 85, False),
                          (DACPins.AO_EN2.value, 85, True),
                          (DACPins.AO_EN3.value, 85, False),
                          (DACPins.AO_EN4.value, 85, True),
                          (DACPins.AO_EN5.value, 85, False),
                          (DACPins.AO_EN6.value, 85, True),
                          (DACPins.AO_EN7.value, 85, False),
                          (DACPins.AO_EN8.value, 85, True)])
def test_read_expander_pin_state(mocker, pin_name, mock_value, result):
    mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__read_register",
                  return_value = mock_value)
    gpio_ctrl = EdgePiGPIOExpander()
    assert gpio_ctrl.read_expander_pin(pin_name) == result

@pytest.mark.parametrize("pin_name, mock_value, result",
                         [(DACPins.AO_EN1.value, 170, True),
                          (DACPins.AO_EN2.value, 170, False),
                          (DACPins.AO_EN3.value, 170, True),
                          (DACPins.AO_EN4.value, 170, False),
                          (DACPins.AO_EN5.value, 170, True),
                          (DACPins.AO_EN6.value, 170, False),
                          (DACPins.AO_EN7.value, 170, True),
                          (DACPins.AO_EN8.value, 170, False),
                          (DACPins.AO_EN1.value, 85, False),
                          (DACPins.AO_EN2.value, 85, True),
                          (DACPins.AO_EN3.value, 85, False),
                          (DACPins.AO_EN4.value, 85, True),
                          (DACPins.AO_EN5.value, 85, False),
                          (DACPins.AO_EN6.value, 85, True),
                          (DACPins.AO_EN7.value, 85, False),
                          (DACPins.AO_EN8.value, 85, True)])
def test_get_expander_pin_direction(mocker, pin_name, mock_value, result):
    mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__read_register",
                  return_value = mock_value)
    gpio_ctrl = EdgePiGPIOExpander()
    assert gpio_ctrl.get_expander_pin_direction(pin_name) == result

@pytest.mark.parametrize("pin_name, mock_value",
                         [(DACPins.AO_EN4.value, 0)])
def test_clear_expander_pin(mocker, pin_name, mock_value):
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__apply_code_to_register")
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__read_register",
        return_value = mock_value)
    gpio_ctrl = EdgePiGPIOExpander()
    prev_pin_config = deepcopy(gpio_ctrl.expander_pin_dict[pin_name])
    gpio_ctrl.clear_expander_pin(pin_name)
    assert prev_pin_config != gpio_ctrl.expander_pin_dict[pin_name]

@pytest.mark.parametrize("pin_name, mock_value",
                         [(DACPins.AO_EN4.value, 0)])
def test_set_expander_pin(mocker, pin_name, mock_value, mock_i2c):
    mock_i2c.return_value = None
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__apply_code_to_register")
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__read_register",
        return_value = mock_value)
    set_pin_dir = mocker.patch(
                        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.set_expander_pin_direction_out")
    gpio_ctrl = EdgePiGPIOExpander()
    prev_pin_config = deepcopy(gpio_ctrl.expander_pin_dict[pin_name])
    gpio_ctrl.set_expander_pin(pin_name)
    if set_pin_dir.call_count == 1:
        gpio_ctrl.expander_pin_dict[pin_name].is_out = True
    assert prev_pin_config != gpio_ctrl.expander_pin_dict[pin_name]

@pytest.mark.parametrize("pin_name, mock_value",
                         [(DACPins.AO_EN4.value, 0)])
def test_set_pin_direction_out(mocker, pin_name, mock_value, mock_i2c):
    mock_i2c.return_value = None
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__apply_code_to_register")
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__read_register",
        return_value = mock_value)
    clear_pin = mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.clear_expander_pin")
    gpio_ctrl = EdgePiGPIOExpander()
    prev_pin_config = deepcopy(gpio_ctrl.expander_pin_dict[pin_name])
    gpio_ctrl.set_expander_pin_direction_out(pin_name)
    if clear_pin.call_count == 1:
        gpio_ctrl.expander_pin_dict[pin_name].is_high = False
    assert prev_pin_config != gpio_ctrl.expander_pin_dict[pin_name]

@pytest.mark.parametrize("pin_name, mock_value",
                         [(DACPins.AO_EN4.value, 0)])
def test_set_pin_direction_in(mocker, pin_name, mock_value, mock_i2c):
    mock_i2c.return_value = None
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__apply_code_to_register")
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__read_register",
        return_value = mock_value)
    gpio_ctrl = EdgePiGPIOExpander()
    prev_pin_config = deepcopy(gpio_ctrl.expander_pin_dict[pin_name])
    gpio_ctrl.set_expander_pin_direction_in(pin_name)
    assert prev_pin_config != gpio_ctrl.expander_pin_dict[pin_name]

@pytest.mark.parametrize("pin_name, mock_value, result",
                         [(DACPins.AO_EN1.value, 170, True),
                          (DACPins.AO_EN2.value, 170, False),
                          (DACPins.AO_EN3.value, 170, True),
                          (DACPins.AO_EN4.value, 170, False),
                          (DACPins.AO_EN5.value, 170, True),
                          (DACPins.AO_EN6.value, 170, False),
                          (DACPins.AO_EN7.value, 170, True),
                          (DACPins.AO_EN8.value, 170, False),
                          (DACPins.AO_EN1.value, 85, False),
                          (DACPins.AO_EN2.value, 85, True),
                          (DACPins.AO_EN3.value, 85, False),
                          (DACPins.AO_EN4.value, 85, True),
                          (DACPins.AO_EN5.value, 85, False),
                          (DACPins.AO_EN6.value, 85, True),
                          (DACPins.AO_EN7.value, 85, False),
                          (DACPins.AO_EN8.value, 85, True)])
def test_toggle_expander_pin(mocker, pin_name, mock_value, result):
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__apply_code_to_register")
    mocker.patch(
        "edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander._EdgePiGPIOExpander__read_register",
        return_value = mock_value)
    gpio_ctrl = EdgePiGPIOExpander()
    gpio_ctrl.toggle_expander_pin(pin_name)
    assert gpio_ctrl.expander_pin_dict[pin_name] != result



""" Command class and methods for DAC devices """

import logging
from bitstring import Bits, pack
from edgepi.dac.dac_constants import (
    NUM_PINS,
    READ_WRITE_SIZE,
    DACChannel as CH,
    EdgePiDacCom as COMMAND,
    EdgePiDacCalibrationConstants as CALIB_CONSTS,
)


_logger = logging.getLogger(__name__)


class DACCommands:
    """Class for representing DAC device commands"""

    def __init__(self, dict_calib_param):
        _logger.info("Initializing DAC Methods")
        self.dict_calib_param = dict_calib_param

    def generate_write_and_update_command(self, ch: int, data: int) -> list:
        """Construct a write and update command"""
        self.check_range(ch, 0, len(CH))
        self.check_range(data, 0, CALIB_CONSTS.RANGE.value)
        return self.combine_command(COMMAND.COM_WRITE_UPDATE.value, CH(ch).value, data)

    def __voltage_to_float_code(self, ch: int, expected: float, dac_gain: int = 1):
        """Convert a voltage to full precision binary code value"""
        float_code =((expected / dac_gain) - self.dict_calib_param[ch].offset) * \
                    (CALIB_CONSTS.RANGE.value / \
                    (CALIB_CONSTS.V_RANGE.value * self.dict_calib_param[ch].gain))
        _logger.debug(f"Full code generated {float_code}")
        return float_code

    def voltage_to_code(self, ch: int, expected: float, dac_gain: int = 1) -> int:
        """
        Convert a voltage to binary code value

        Args:
            ch (int): the DAC channel to write voltage to (0-indexed)

            expected (float): the voltage input by the user

            dac_gain (int): dac gain state, x1 or x2, when x2 voltage range extened to 10V from 5V

        Returns:
            int: 16 bit binary code value for writing voltage value to DAC
        """
        # DAC channels are 0 indexed
        self.check_range(ch, 0, NUM_PINS - 1)
        float_code = self.__voltage_to_float_code(ch, expected, dac_gain)
        _logger.debug(f"Int code generated {int(float_code)}")
        # DAC only accepts int values, round to nearest int
        return round(float_code)

    @staticmethod
    def extract_read_data(read_code: list) -> int:
        """
        Extracts bits corresponding to voltage code from a list containing
        the byte values of a DAC register read.

        Args:
            read_code (list): a list of unsigned int byte values from DAC read

        Returns:
            int: the unsigned int value of the 16 bits corresponding to voltage code

        Raises:
            ValueError: if read_code does not contain exactly 3 byte values
        """
        if len(read_code) != READ_WRITE_SIZE:
            raise ValueError("code must contain exactly 3 byte values")

        bits = pack("uint:8, uint:8, uint:8", read_code[0], read_code[1], read_code[2])

        # B23 to DB20 contain undefined data, and the last 16 bits contain the
        # DB19 to DB4 DAC register contents. B23 (MSB) is index 0 here.
        return bits[-16:].uint

    def __code_to_float_voltage(self, ch: int, code: int, dac_gain: int = 1) -> float:
        """Convert a voltage to float voltage"""
        voltage = (CALIB_CONSTS.V_RANGE.value /
                   CALIB_CONSTS.RANGE.value * self.dict_calib_param[ch].gain) *\
                  (code*dac_gain) + self.dict_calib_param[ch].offset
        return voltage

    def code_to_voltage(self, ch: int, code: int, dac_gain: int = 1) -> float:
        """
        Convert a 16 bit binary code value to voltage

        Args:
            ch (int): the DAC channel the code was obtained from (0-indexed)

            code (int): 16 bit unsigned int value of a DAC register read

            dac_gain (int): dac gain state, x1 or x2, when x2 voltage range extened to 10V from 5V

        Returns:
            float: voltage corresponding to 16 bit binary code
        """
        voltage = self.__code_to_float_voltage(ch, code, dac_gain)
        _logger.debug(f"code_to_voltage: code {hex(code)} = {voltage} V")
        return voltage

    @staticmethod
    # pylint: disable=inconsistent-return-statements
    def combine_command(op_code: int, ch: int, value: int) -> list:
        """
        Combine op_code, channel and value into a message frame to send out to DAC device

        Args:
            op_code (COMMAND.value): AD5675 DAC command

            ch (CH.value): AD5657 DAC channel address

            value: 16 bit data

        Returns:
            list: a list of bytes that form a DAC input shift register message frame,
                i.e. of the form [op_code_byte, ch_byte, value_byte]
        """
        DACCommands.check_for_int([op_code, ch, value])
        temp = (op_code << 20) + (ch << 16) + value
        combined_cmnd = [temp >> 16, (temp >> 8) & 0xFF, temp & 0xFF]
        _logger.debug(f"Combined Command is: {combined_cmnd}")
        return combined_cmnd

    @staticmethod
    def check_for_int(target_list: list) -> bool:
        """Checks if a list contains only integer values"""
        if all(isinstance(value, int) for value in target_list) and target_list:
            return True

        raise ValueError(f"Non integer value passed{target_list}")

    @staticmethod
    def check_range(target, range_min, range_max) -> bool:
        """Validates target is in range between a min and max value"""
        if range_min <= target <= range_max:
            return True

        raise ValueError(f"Target {target} is out of range ")

    @staticmethod
    def generate_power_code(dac_state: list) -> int:
        """
        Converts a list containing the EdgePi DAC's power state for each channel
        to a binary code value.

        Args:
            dac_state (list): a list whose entries represent the power mode of each DAC
                channel, ordered sequentially from DAC 7 (index 0) to DAC 0 (index 7).

        Returns:
            int: a 16 bit binary code value for updating DAC channel power mode
        """
        data = Bits()
        for value in dac_state:
            data += Bits(uint=value, length=2)
        return data.uint



"""unit tests for edgepi_gpio module"""

# pylint: disable=protected-access
# pylint: disable=C0413

from unittest import mock
import sys
if sys.platform != 'linux':
    sys.modules['periphery'] = mock.MagicMock()

from contextlib import nullcontext as does_not_raise
import pytest
from edgepi.gpio.gpio_configs import generate_expander_pin_info, generate_gpiochip_pin_info
from edgepi.gpio.gpio_constants import GpioPins
from edgepi.gpio.edgepi_gpio import EdgePiGPIO, PinNameNoneError, PinNameNotFound

def test_edgepi_gpio_init(mocker):
    mocker.patch('edgepi.peripherals.i2c.I2C')
    mocker.patch('edgepi.peripherals.gpio.GPIO')
    edgepi_gpio = EdgePiGPIO()
    expander_pin_dict = generate_expander_pin_info()
    gpiochip_pin_dict = generate_gpiochip_pin_info()
    assert edgepi_gpio.expander_pin_dict == expander_pin_dict
    assert edgepi_gpio.gpiochip_pins_dict == gpiochip_pin_dict

@pytest.mark.parametrize("pin_name, error",
                        [(GpioPins.AO_EN1.value, does_not_raise()),
                         (GpioPins.DIN1.value, does_not_raise()),
                         (None, pytest.raises(PinNameNoneError)),
                         ("Does not exits", pytest.raises(PinNameNotFound))])
def test_edgepi__pin_name_check(pin_name, error):
    edgepi_gpio = EdgePiGPIO()
    with error:
        edgepi_gpio._EdgePiGPIO__pin_name_check(pin_name)

@pytest.mark.parametrize("pin_name, mock_value, result, error",
                        [(GpioPins.AO_EN1.value,[True, None], True, does_not_raise()),
                         (GpioPins.DIN1.value,[None, True], True, does_not_raise()),
                         (None, [None, None], None, pytest.raises(PinNameNoneError)),
                         ("Does not exits",[None, None], None, pytest.raises(PinNameNotFound))])
def test_edgepi_gpio_read_pin_state(mocker, pin_name, mock_value, result, error):
    mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOChip.read_gpio_pin_state',
                  return_value = mock_value[1])
    mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.read_expander_pin",
                  return_value = mock_value[0])
    edgepi_gpio = EdgePiGPIO()
    with error:
        state = edgepi_gpio.read_pin_state(pin_name)
        assert state == result

@pytest.mark.parametrize("pin_name, mock_value, error",
                        [(GpioPins.DIN1.value,[None, True], does_not_raise()),
                         (None, [None, None], pytest.raises(PinNameNoneError)),
                         ("Does not exits",[None, None], pytest.raises(PinNameNotFound))])
def test_edgepi_gpio_set_pin_state(mocker, pin_name, mock_value, error):
    mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOChip.write_gpio_pin_state',
                  return_value = mock_value[1])
    mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.set_expander_pin",
                  return_value = mock_value[0])
    edgepi_gpio = EdgePiGPIO()
    with error:
        edgepi_gpio.set_pin_state(pin_name)

@pytest.mark.parametrize("pin_name, mock_value, error",
                        [(GpioPins.DIN1.value,[None, True], does_not_raise()),
                         (None, [None, None], pytest.raises(PinNameNoneError)),
                         ("Does not exits",[None, None], pytest.raises(PinNameNotFound))])
def test_edgepi_gpio_clear_pin_state(mocker, pin_name, mock_value, error):
    mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOChip.write_gpio_pin_state',
                  return_value = mock_value[1])
    mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.clear_expander_pin",
                  return_value = mock_value[0])
    edgepi_gpio = EdgePiGPIO()
    with error:
        edgepi_gpio.set_pin_state(pin_name)

@pytest.mark.parametrize("pin_name, mock_value, result, error",
                        [(GpioPins.DIN1.value,[None, None], True, does_not_raise()),
                         (GpioPins.DOUT1.value,[None, None], False, does_not_raise()),
                         (GpioPins.AO_EN1.value,[None, False], False, does_not_raise()),
                         (None, [None, None], None, pytest.raises(PinNameNoneError)),
                         ("Does not exits",[None, None], None, pytest.raises(PinNameNotFound))])
def test_edgepi_gpio_get_pin_direction(mocker, pin_name, mock_value, result, error):
    mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.get_expander_pin_direction',
                  return_value = mock_value[1])
    edgepi_gpio = EdgePiGPIO()
    with error:
        direction = edgepi_gpio.get_pin_direction(pin_name)
        assert direction == result

@pytest.mark.parametrize("pin_name, error",
                        [(GpioPins.DIN1.value, does_not_raise()),
                         (GpioPins.DOUT1.value, does_not_raise()),
                         (GpioPins.AO_EN1.value, does_not_raise()),
                         (None, pytest.raises(PinNameNoneError)),
                         ("Does not exits", pytest.raises(PinNameNotFound))])
def test_edgepi_gpio_set_pin_direction_in(mocker, pin_name, error):
    exp = mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.set_expander_pin_direction_in')
    gpio = mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOChip.set_gpio_pin_dir')
    edgepi_gpio = EdgePiGPIO()
    with error:
        edgepi_gpio.set_pin_direction_in(pin_name)
        # pylint: disable = expression-not-assigned
        exp.assert_called_once_with(pin_name) if pin_name in edgepi_gpio.expander_pin_dict else \
            gpio.assert_called_once_with(pin_name, True)

@pytest.mark.parametrize("pin_name, error",
                        [(GpioPins.DIN1.value, does_not_raise()),
                         (GpioPins.DOUT1.value, does_not_raise()),
                         (GpioPins.AO_EN1.value, does_not_raise()),
                         (None, pytest.raises(PinNameNoneError)),
                         ("Does not exits", pytest.raises(PinNameNotFound))])
def test_edgepi_gpio_set_pin_direction_out(mocker, pin_name, error):
    exp = mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.set_expander_pin_direction_out')
    gpio = mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOChip.set_gpio_pin_dir')
    edgepi_gpio = EdgePiGPIO()
    with error:
        edgepi_gpio.set_pin_direction_out(pin_name)
        # pylint: disable = expression-not-assigned
        exp.assert_called_once_with(pin_name) if pin_name in edgepi_gpio.expander_pin_dict else \
            gpio.assert_called_once_with(pin_name, False)

@pytest.mark.parametrize("pin_name, error",
                        [(GpioPins.DIN1.value, does_not_raise()),
                         (GpioPins.DOUT1.value, does_not_raise()),
                         (GpioPins.AO_EN1.value, does_not_raise()),
                         (None, pytest.raises(PinNameNoneError)),
                         ("Does not exits", pytest.raises(PinNameNotFound))])
def test_edgepi_gpio_toggle_pin(mocker, pin_name, error):
    exp = mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOExpander.toggle_expander_pin')
    gpio = mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIOChip.toggle_gpio_pin_state')
    edgepi_gpio = EdgePiGPIO()
    with error:
        edgepi_gpio.toggle_pin(pin_name)
        # pylint: disable = expression-not-assigned
        exp.assert_called_once_with(pin_name) if pin_name in edgepi_gpio.expander_pin_dict else \
            gpio.assert_called_once_with(pin_name)



""" Constants for DAC device modules """


from enum import Enum


DAC_PRECISION = 3  # decimal place precision for voltage conversions
SW_RESET = 0x1234  # software reset command data bits
NULL_BITS = 0x0
READ_WRITE_SIZE = 3  # size of DAC read/write in bytes
UPPER_LIMIT = 5.0 # upper limit for voltage writes to DAC


class EdgePiDacCom(Enum):
    """Commands for EdgePi DAC"""

    COM_NOP = 0x0
    COM_WRITE_INPUT = 0x1  # LDAC pin is held low, so not used
    COM_UPDATE_DAC = 0x2
    COM_WRITE_UPDATE = 0x3
    COM_POWER_DOWN_OP = 0x4
    COM_HW_LDAC_MASK = 0x5  # LDAC pin is held low, so not used
    COM_SW_RESET = 0x6  # resets to POR, address bits set to 0x0 and data bits set to 0x1234
    COM_GAIN = 0x7
    COM_DCEN = 0x8  # Daisy chain setup enable
    COM_READBACK = 0x9  # used for daisy chain operation
    COM_UPDATE_ALL_CH = (
        0x0A  # update all channels of the input register simultaneously with the input data
    )
    # update all channel of teh input register and DAC register simultaneously with the input data
    COM_UPDATE_ALL_DAC = 0x0B


class DACChannel(Enum):
    """EdgePi DAC channel addresses"""

    AOUT0 = 0x0
    AOUT1 = 0x1
    AOUT2 = 0x2
    AOUT3 = 0x3
    AOUT4 = 0x4
    AOUT5 = 0x5
    AOUT6 = 0x6
    AOUT7 = 0x7


NUM_PINS = len(DACChannel)


class EdgePiDacCalibrationConstants(Enum):
    """EdgePi DAC calibration constants"""

    V_RANGE = 5.0000
    RANGE = 65535
    DAC_GAIN_FACTOR = 2


class PowerMode(Enum):
    """
    EdgePi DAC power modes

    Attributes:
        NORMAL: normal power consumption of 1.1 mA typically

        POWER_DOWN_GROUND: low power consumption, 1 μA typically. DAC output stage
            connected internally to GND through 1 kΩ resistor.

        POWER_DOWN_3_STATE: low power consumption, 1 μA typically. DAC output stage
            left open circuited (three-state).
    """

    NORMAL = 0x0
    POWER_DOWN_GROUND = 0x1
    POWER_DOWN_3_STATE = 0x3


class AOPins(Enum):
    """Analog/Digital out gpio pin names"""

    AO_EN1 = "AO_EN1"
    AO_EN2 = "AO_EN2"
    AO_EN3 = "AO_EN3"
    AO_EN4 = "AO_EN4"
    AO_EN5 = "AO_EN5"
    AO_EN6 = "AO_EN6"
    AO_EN7 = "AO_EN7"
    AO_EN8 = "AO_EN8"

class GainPin(Enum):
    """DAC gain enable/disable pin"""
    DAC_GAIN = "DAC_GAIN"



"""
Module for I2C devices

Classes:
    I2CDevice
"""

from typing import Union
from periphery import I2C

class I2CDevice():
    '''
    I2C Device class
    '''
    def __init__(self, fd: str = None):
        self.fd = fd
        self.i2cdev = I2C(fd)

    def set_read_msg(self, addr:Union[int,list] = None, msg:list = None):
        '''
        set Read message to be sent through I2C.
        Attributes:
            addr(int or list): Register address to read from
            Msg: list of place holder bytes
        Return:
            MsgList: list of I2C.Message() objects containing msg to be sent
        '''
        list_msg = [self.i2cdev.Message([addr], read = False),
                    self.i2cdev.Message(msg, read = True)] \
                   if isinstance(addr, int) else \
                   [self.i2cdev.Message(addr, read = False),
                    self.i2cdev.Message(msg, read = True)]
        return list_msg

    def set_write_msg(self, addr:Union[int,list] = None, msg:list = None):
        '''
        set Write message to be sent through I2C.
        Attributes:
            addr(int or list): Register address to write to
            Msg: list of Msg bytes
        Return:
            MsgList: list of I2C.Message() objects containing msg to be sent
        '''
        list_msg = [self.i2cdev.Message([addr]+msg, read = False)] if isinstance(addr, int) else \
                   [self.i2cdev.Message(addr+msg, read = False)]
        return list_msg

    def transfer(self, dev_addr: int = None, msg:list = None):
        '''
        Message Transfer
        Attributes:
            dev_addr: hardware device address
            Msg: list of Message class objects
        Return:
            MsgList: list of message bytes if reading flag was set
        '''
        self.i2cdev.transfer(dev_addr, msg)
        if len(msg)>1:
            return msg[1].data
        return None

    def close(self):
        '''
        Close I2C device
        '''
        self.i2cdev.close()



"""ADC voltage reading configuration and utility methods"""


import logging

from bitstring import BitArray
from edgepi.adc.adc_constants import ADCReadInfo
from edgepi.calibration.calibration_constants import CalibParam
from edgepi.utilities.utilities import bitstring_from_list
from edgepi.adc.adc_crc_8_atm import CRC_8_ATM_LUT


# TODO: retrieve these values from EEPROM once added
STEP_DOWN_RESISTOR_1 = 19.1
STEP_DOWN_RESISTOR_2 = 4.99
REFERENCE_VOLTAGE = 2.5


_logger = logging.getLogger(__name__)


def _is_negative_voltage(code: BitArray):
    """
    Determines if voltage code is negative value
    """
    return code[0] == 1


def _code_to_input_voltage(code: int, v_ref: float, num_bits: int):
    """
    Converts digital code obtained from ADC voltage read to
    ADC input voltage (i.e. voltage measured at ADC) based on voltage range
    provided by reference voltage

    Args:
        `code` (int): uint value of ADC voltage read bytes

        `v_ref` (float): ADC reference voltage in Volts

        `num_bits` (int): number of bits in ADC voltage read (24 or 32)
    """
    voltage_range = v_ref / 2 ** (num_bits - 1)
    return float(code) * voltage_range


def _adc_voltage_to_input_voltage(v_in: float, gain: float, offset: float):
    """
    Converts ADC input voltage (i.e. voltage measured at ADC) to
    ADC output voltage (i.e. voltage measured at terminal block)
    """
    step_up_ratio = (STEP_DOWN_RESISTOR_1 + STEP_DOWN_RESISTOR_2) / STEP_DOWN_RESISTOR_2
    return v_in * step_up_ratio * gain - offset


def code_to_voltage(code: list[int], adc_info: ADCReadInfo, calibs: CalibParam) -> float:
    """
    Converts ADC voltage read digital code to output voltage (voltage measured at terminal block)

    Args:
        `code` (list[int]): code bytes retrieved from ADC voltage read

        `adc_info` (ADCReadInfo): data about this adc's voltage reading configuration

        `calibs` (CalibParam): voltage reading gain and offset calibration values

    Returns:
        `float`: voltage value (V) corresponding to `code`
    """
    code_bits = bitstring_from_list(code)
    num_bits = adc_info.num_data_bytes * 8

    # code is given in 2's complement, remove leading 1
    if _is_negative_voltage(code_bits):
        code_bits[0] = 0

    v_in = _code_to_input_voltage(code_bits.uint, REFERENCE_VOLTAGE, num_bits)

    v_out = _adc_voltage_to_input_voltage(v_in, calibs.gain, calibs.offset)

    return v_out


def code_to_temperature(
    code: list[int],
    ref_resistance: float,
    rtd_sensor_resistance: float,
    rtd_sensor_resistance_variation: float
    ) -> float:
    """
    Converts ADC voltage read digital code to temperature. Intended for use in RTD sampling.

    Args:
        `code` (list[int]): code bytes retrieved from ADC voltage read

       `ref_resistance` (float): EdgePi-specific RTD reference resistance (Ohms)

       `rtd_sensor_resistance` (float): RTD material-dependent resistance value (Ohms)

       `rtd_sensor_resistance_variation` (float): RTD model-dependent resistance variation (Ohms/°C)

    Returns:
        `float`: temperature value (°C) corresponding to `code`
    """
    code_bits = bitstring_from_list(code)

    # refer to https://github.com/osensa/edgepi-python-sdk/issues/159 for computation details
    r_rtd = code_bits.uint / (2 ** 30) * ref_resistance

    temperature = (r_rtd - rtd_sensor_resistance) / rtd_sensor_resistance_variation
    _logger.debug(f"computed rtd temperature = {temperature}, from code = {code_bits.uint}")

    return temperature


class CRCCheckError(Exception):
    """Raised if CRC check fails to match CRC code generated by ADS1263"""


# use this with `generate_crc_8_table` below to get LUT for CRC-8-ATM polynomial
CRC_8_ATM_GEN = 0x07


def generate_crc_8_table(crc_polynomial: int):
    """
    Generates LUT containing CRC-8 values for all possible byte (8 bit) values, i.e.
    every entry is the CRC-8 value for the ith byte value from 0,...,255.

    Adapted from:
    https://github.com/osensa/rtu-adc/blob/6fac17efb955c2e8bb40740dd46e48ae51e1d5ff/crc8atm.py

    Args:
        `crc_polynomial`: the CRC-8 polynomial used to generate this LUT

    Returns:
        `list`: the LUT corresponding to this CRC-8 polynomial
    """
    table = list(range(256))

    # for every byte value from 0 to 255
    for index, crc in enumerate(table):
        for _ in range(8):
            # scan bitwise through byte to align leading 1 with CRC polynomial
            # perform divison (XOR) operation once aligned
            crc = (crc << 1) ^ (crc_polynomial if (crc & 0x80) else 0)

        # store only low byte of computed value to account for left shifting
        table[index] = crc & 0xFF

    return table


def check_crc(voltage_bytes: list, crc_code: int):
    """
    Performs CRC-8 check on a list of data bytes. By default uses
    CRC-8-ATM polynomial table

    Adapted from:
    https://github.com/osensa/rtu-adc/blob/6fac17efb955c2e8bb40740dd46e48ae51e1d5ff/crc8atm.py

    Args:
        `data_bytes` (list): a list of bytes ordered as
            [data_1, data_2,..., data_n, crc_8]

        `crc_table`: a previously generated LUT for the CRC-8-ATM
            polynomial
    """
    crc = 0

    data_bytes = voltage_bytes + [crc_code]

    # compute CRC value of data bytes byte wise using LUT
    # which holds CRC values for all possible byte values
    for byte in data_bytes:
        index = crc ^ byte
        crc = CRC_8_ATM_LUT[index]

    if crc != 0:
        raise CRCCheckError(
            f"CRC check failed with codes: computed={hex(crc)}, expected={hex(crc_code)}"
        )



"""Unit tests for adc_voltage.py module"""


from contextlib import nullcontext as does_not_raise

import pytest
from edgepi.adc.adc_constants import ADCNum
from edgepi.adc.adc_voltage import (
    check_crc,
    CRCCheckError,
    _code_to_input_voltage,
    generate_crc_8_table,
    code_to_temperature,
    CRC_8_ATM_GEN
)
from edgepi.adc.adc_crc_8_atm import CRC_8_ATM_LUT

# pylint: disable=too-many-lines

# mock calib values
V_REF = 2.5
GAIN = 1
OFFSET = 0


@pytest.mark.parametrize(
    "code, voltage, num_bytes",
    [
        (0x00000000, 0, ADCNum.ADC_1.value.num_data_bytes),
        (0x00000000, 0, ADCNum.ADC_2.value.num_data_bytes),
        # based on a reference voltage of 2.5 V
        (
            0xFFFFFFFF,
            5.0,
            ADCNum.ADC_1.value.num_data_bytes,
        ),
        (
            0xFFFFFF,
            5.0,
            ADCNum.ADC_2.value.num_data_bytes,
        ),
    ],
)
def test_code_to_voltage(code, voltage, num_bytes):
    assert pytest.approx(_code_to_input_voltage(code, V_REF, num_bytes * 8)) == voltage


@pytest.mark.parametrize(
    "voltage_bytes, crc_code, err",
    [
        ([51, 16, 126, 166], 62, does_not_raise()),
        ([51, 14, 170, 195], 98, does_not_raise()),
        ([51, 16, 133, 237], 75, does_not_raise()),
        ([51, 17, 166, 166], 71, does_not_raise()),
        ([51, 16, 148, 157], 94, does_not_raise()),
        ([51, 14, 144, 155], 150, does_not_raise()),
        ([51, 14, 166, 18], 167, does_not_raise()),
        ([51, 16, 5, 109], 116, does_not_raise()),
        ([51, 15, 16, 130], 4, does_not_raise()),
        ([51, 16, 126, 166], 61, pytest.raises(CRCCheckError)),
        ([51, 14, 170, 195], 99, pytest.raises(CRCCheckError)),
        ([51, 16, 133, 237], 70, pytest.raises(CRCCheckError)),
    ],
)
def test_crc_8_atm_adc_1(voltage_bytes, crc_code, err):
    with err:
        check_crc(voltage_bytes, crc_code)


def test_generate_crc_8_table():
    assert generate_crc_8_table(CRC_8_ATM_GEN) == CRC_8_ATM_LUT


@pytest.mark.parametrize(
    "code, ref_resistance, temp_offset, rtd_conv_constant",
    [
        ([51, 16, 126, 166], 1326.20, 100, 0.385),
        ([0x8, 0x43, 0x1C, 0x45], 1326.20, 100, 0.385),
    ]
)
def test_code_to_temperature(code, ref_resistance, temp_offset, rtd_conv_constant):
    # TODO: add check for expected value later if any values are known. No errors raised
    # is good enough for now.
    code_to_temperature(code, ref_resistance, temp_offset, rtd_conv_constant)



"""Utility module for computing ADC conversion delay time"""

# ADC conversion delay time is affected by the following factors:
# - Conversion delay (REG_MODE0) -> currently not user-configurable
# - Digital filter mode (REG_MODE1)
# - Data rate mode (REG_MODE2)
# - IDAC/Chop mode (REG_MODE0) -> currently not user-configurable


from edgepi.adc.adc_constants import (
    FilterMode as FILT,
    ADC1DataRate as DR1,
    ADC2DataRate as DR2,
    ADCNum,
)


# pulse conversion delay times for ADC1, or first conversion delay time for continuous conversions
ADC1_INITIAL_DELAYS = {
    DR1.SPS_2P5.value.op_code: {
        FILT.SINC1.value.op_code: 400.4,
        FILT.SINC2.value.op_code: 800.4,
        FILT.SINC3.value.op_code: 1200.0,
        FILT.SINC4.value.op_code: 1600.0,
        FILT.FIR.value.op_code: 402.4,
    },
    DR1.SPS_5.value.op_code: {
        FILT.SINC1.value.op_code: 200.4,
        FILT.SINC2.value.op_code: 400.4,
        FILT.SINC3.value.op_code: 600.4,
        FILT.SINC4.value.op_code: 800.4,
        FILT.FIR.value.op_code: 202.2,
    },
    DR1.SPS_10.value.op_code: {
        FILT.SINC1.value.op_code: 100.4,
        FILT.SINC2.value.op_code: 200.4,
        FILT.SINC3.value.op_code: 300.4,
        FILT.SINC4.value.op_code: 400.4,
        FILT.FIR.value.op_code: 102.2,
    },
    DR1.SPS_16P6.value.op_code: {
        FILT.SINC1.value.op_code: 60.35,
        FILT.SINC2.value.op_code: 120.4,
        FILT.SINC3.value.op_code: 180.4,
        FILT.SINC4.value.op_code: 240.4,
        FILT.FIR.value.op_code: 60.35,
    },
    DR1.SPS_20.value.op_code: {
        FILT.SINC1.value.op_code: 50.35,
        FILT.SINC2.value.op_code: 100.4,
        FILT.SINC3.value.op_code: 150.4,
        FILT.SINC4.value.op_code: 200.4,
        FILT.FIR.value.op_code: 52.22,
    },
    DR1.SPS_50.value.op_code: {
        FILT.SINC1.value.op_code: 20.35,
        FILT.SINC2.value.op_code: 40.42,
        FILT.SINC3.value.op_code: 60.42,
        FILT.SINC4.value.op_code: 80.42,
        FILT.FIR.value.op_code: 20.35,
    },
    DR1.SPS_60.value.op_code: {
        FILT.SINC1.value.op_code: 17.02,
        FILT.SINC2.value.op_code: 33.76,
        FILT.SINC3.value.op_code: 50.42,
        FILT.SINC4.value.op_code: 67.09,
        FILT.FIR.value.op_code: 17.02,
    },
    DR1.SPS_100.value.op_code: {
        FILT.SINC1.value.op_code: 10.35,
        FILT.SINC2.value.op_code: 20.42,
        FILT.SINC3.value.op_code: 30.42,
        FILT.SINC4.value.op_code: 40.42,
        FILT.FIR.value.op_code: 10.35,
    },
    DR1.SPS_400.value.op_code: {
        FILT.SINC1.value.op_code: 2.855,
        FILT.SINC2.value.op_code: 5.424,
        FILT.SINC3.value.op_code: 7.924,
        FILT.SINC4.value.op_code: 10.42,
        FILT.FIR.value.op_code: 2.855,
    },
    DR1.SPS_1200.value.op_code: {
        FILT.SINC1.value.op_code: 1.188,
        FILT.SINC2.value.op_code: 2.091,
        FILT.SINC3.value.op_code: 2.924,
        FILT.SINC4.value.op_code: 3.758,
        FILT.FIR.value.op_code: 1.188,
    },
    DR1.SPS_2400.value.op_code: {
        FILT.SINC1.value.op_code: 0.771,
        FILT.SINC2.value.op_code: 1.258,
        FILT.SINC3.value.op_code: 1.674,
        FILT.SINC4.value.op_code: 2.091,
        FILT.FIR.value.op_code: 0.771,
    },
    DR1.SPS_4800.value.op_code: {
        FILT.SINC1.value.op_code: 0.563,
        FILT.SINC2.value.op_code: 0.8409,
        FILT.SINC3.value.op_code: 1.049,
        FILT.SINC4.value.op_code: 1.258,
        FILT.FIR.value.op_code: 0.563,
    },
    DR1.SPS_7200.value.op_code: {
        FILT.SINC1.value.op_code: 0.494,
        FILT.SINC2.value.op_code: 0.702,
        FILT.SINC3.value.op_code: 0.841,
        FILT.SINC4.value.op_code: 0.980,
        FILT.FIR.value.op_code: 0.494,
    },
    DR1.SPS_14400.value.op_code: {
        FILT.SINC1.value.op_code: 0.424,
        FILT.SINC2.value.op_code: 0.424,
        FILT.SINC3.value.op_code: 0.424,
        FILT.SINC4.value.op_code: 0.424,
        FILT.FIR.value.op_code: 0.424,
    },
    DR1.SPS_19200.value.op_code: {
        FILT.SINC1.value.op_code: 0.337,
        FILT.SINC2.value.op_code: 0.337,
        FILT.SINC3.value.op_code: 0.337,
        FILT.SINC4.value.op_code: 0.337,
        FILT.FIR.value.op_code: 0.337,
    },
    DR1.SPS_38400.value.op_code: {
        FILT.SINC1.value.op_code: 0.207,
        FILT.SINC2.value.op_code: 0.207,
        FILT.SINC3.value.op_code: 0.207,
        FILT.SINC4.value.op_code: 0.207,
        FILT.FIR.value.op_code: 0.207,
    },
}


# ADC1 continuous conversion delay times, for conversions 2..n (following first conversion)
ADC1_CONT_DELAYS = {
    DR1.SPS_2P5.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_2P5.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_5.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_5.value.op_code][FILT.SINC1.value.op_code],
    DR1.SPS_10.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_10.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_16P6.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_16P6.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_20.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_20.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_50.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_50.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_60.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_60.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_100.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_100.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_400.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_400.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_1200.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_1200.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_2400.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_2400.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_4800.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_4800.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_7200.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_7200.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_14400.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_14400.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_19200.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_19200.value.op_code][
        FILT.SINC1.value.op_code
    ],
    DR1.SPS_38400.value.op_code: ADC1_INITIAL_DELAYS[DR1.SPS_38400.value.op_code][
        FILT.SINC1.value.op_code
    ],
}


# continuous conversion delay times for ADC2 (ADC2 only uses continuous mode)
ADC2_DELAYS = {
    DR2.SPS_10.value.op_code: 121,
    DR2.SPS_100.value.op_code: 31.2,
    DR2.SPS_400.value.op_code: 8.71,
    DR2.SPS_800.value.op_code: 4.97,
}


def expected_initial_time_delay(adc_num: ADCNum, data_rate: int, filter_mode: int):
    """
    Computes conversion latency (ms) based on ADC configuration
    for PULSE mode reads or initial read in CONTINUOUS mode.

    Args:
        `adc_num` (ADCNum): ADC whose conversions are being sampled

        `data_rate` (int): opcode value of data rate bits

        `filter_mode` (int): opcode value of filter mode bits

    Returns:
        `float`: conversion delay in milliseconds (ms)
    """
    if adc_num == ADCNum.ADC_1:
        # this is the initial delay for both pulse and continuous modes
        return ADC1_INITIAL_DELAYS[data_rate][filter_mode]

    # no initial figures given in documentation, but estimate initial delay
    # is 3 times longer than subsequent conversion delays
    return ADC2_DELAYS[data_rate] * 3


def expected_continuous_time_delay(adc_num: ADCNum, data_rate: int):
    """
    Computes conversion latency (ms) based on ADC configuration
    for reads 2...n in continuous conversion mode.

    Args:
        `adc_num` (ADCNum): ADC whose conversions are being sampled

        `data_rate` (int): opcode value of data rate bits

    Returns:
        `float`: conversion delay in milliseconds (ms)
    """
    if adc_num == ADCNum.ADC_1:
        return ADC1_CONT_DELAYS[data_rate]

    return ADC2_DELAYS[data_rate]



"""unit tests for gpio_constants module"""


import pytest
from edgepi.reg_helper.reg_helper import _apply_opcode
from edgepi.gpio.gpio_constants import (
    GpioAOutputClear,
    GpioAOutputSet,
    GpioAPinDirIn,
    GpioAPinDirOut,
)


@pytest.mark.parametrize(
    "reg_value, opcode, updated_reg_value",
    [
        (0b00000000, GpioAOutputSet.SET_OUTPUT_1.value, 0b00000001),  # Set Bit1
        (0b00000001, GpioAOutputSet.SET_OUTPUT_2.value, 0b00000011),  # Set Bit2
        (0b00000011, GpioAOutputSet.SET_OUTPUT_3.value, 0b00000111),  # Set Bit3
        (0b00000111, GpioAOutputSet.SET_OUTPUT_4.value, 0b00001111),  # Set Bit4
        (0b00001111, GpioAOutputSet.SET_OUTPUT_5.value, 0b00011111),  # Set Bit5
        (0b00011111, GpioAOutputSet.SET_OUTPUT_6.value, 0b00111111),  # Set Bit6
        (0b00111111, GpioAOutputSet.SET_OUTPUT_7.value, 0b01111111),  # Set Bit7
        (0b01111111, GpioAOutputSet.SET_OUTPUT_8.value, 0b11111111),  # Set Bit8
        (0b00000000, GpioAOutputSet.SET_OUTPUT_ALL.value, 0b11111111),  # Set ALL
    ],
)
def test_output_set(reg_value, opcode, updated_reg_value):
    assert _apply_opcode(reg_value, opcode) == updated_reg_value


@pytest.mark.parametrize(
    "reg_value, opcode, updated_reg_value",
    [
        (0b11111111, GpioAOutputClear.CLEAR_OUTPUT_8.value, 0b01111111),  # Claer Bit8
        (0b01111111, GpioAOutputClear.CLEAR_OUTPUT_7.value, 0b00111111),  # Claer Bit7
        (0b00111111, GpioAOutputClear.CLEAR_OUTPUT_6.value, 0b00011111),  # Claer Bit6
        (0b00011111, GpioAOutputClear.CLEAR_OUTPUT_5.value, 0b00001111),  # Claer Bit5
        (0b00001111, GpioAOutputClear.CLEAR_OUTPUT_4.value, 0b00000111),  # Claer Bit4
        (0b00000111, GpioAOutputClear.CLEAR_OUTPUT_3.value, 0b00000011),  # Claer Bit3
        (0b00000011, GpioAOutputClear.CLEAR_OUTPUT_2.value, 0b00000001),  # Claer Bit2
        (0b00000001, GpioAOutputClear.CLEAR_OUTPUT_1.value, 0b00000000),  # Claer Bit1
        (0b11111111, GpioAOutputClear.CLEAR_OUTPUT_ALL.value, 0b00000000),  # Claer ALL
    ],
)
def test_output_clear(reg_value, opcode, updated_reg_value):
    assert _apply_opcode(reg_value, opcode) == updated_reg_value


@pytest.mark.parametrize(
    "reg_value, opcode, updated_reg_value",
    [
        (0b11111111, GpioAPinDirOut.PIN8_DIR_OUT.value, 0b01111111),  # Output Dir Bit8
        (0b01111111, GpioAPinDirOut.PIN7_DIR_OUT.value, 0b00111111),  # Output Dir Bit7
        (0b00111111, GpioAPinDirOut.PIN6_DIR_OUT.value, 0b00011111),  # Output Dir Bit6
        (0b00011111, GpioAPinDirOut.PIN5_DIR_OUT.value, 0b00001111),  # Output Dir Bit5
        (0b00001111, GpioAPinDirOut.PIN4_DIR_OUT.value, 0b00000111),  # Output Dir Bit4
        (0b00000111, GpioAPinDirOut.PIN3_DIR_OUT.value, 0b00000011),  # Output Dir Bit3
        (0b00000011, GpioAPinDirOut.PIN2_DIR_OUT.value, 0b00000001),  # Output Dir Bit2
        (0b00000001, GpioAPinDirOut.PIN1_DIR_OUT.value, 0b00000000),  # Output Dir Bit1
        (0b11111111, GpioAPinDirOut.ALL_DIR_OUT.value, 0b00000000),  # Output Dir ALL
    ],
)
def test_pin_dir_out(reg_value, opcode, updated_reg_value):
    assert _apply_opcode(reg_value, opcode) == updated_reg_value


@pytest.mark.parametrize(
    "reg_value, opcode, updated_reg_value",
    [
        (0b00000000, GpioAPinDirIn.PIN1_DIR_IN.value, 0b00000001),  # Input Dir Bit8
        (0b00000001, GpioAPinDirIn.PIN2_DIR_IN.value, 0b00000011),  # Input Dir Bit7
        (0b00000011, GpioAPinDirIn.PIN3_DIR_IN.value, 0b00000111),  # Input Dir Bit6
        (0b00000111, GpioAPinDirIn.PIN4_DIR_IN.value, 0b00001111),  # Input Dir Bit5
        (0b00001111, GpioAPinDirIn.PIN5_DIR_IN.value, 0b00011111),  # Input Dir Bit4
        (0b00011111, GpioAPinDirIn.PIN6_DIR_IN.value, 0b00111111),  # Input Dir Bit3
        (0b00111111, GpioAPinDirIn.PIN7_DIR_IN.value, 0b01111111),  # Input Dir Bit2
        (0b01111111, GpioAPinDirIn.PIN8_DIR_IN.value, 0b11111111),  # Input Dir Bit1
        (0b00000000, GpioAPinDirIn.ALL_DIR_IN.value, 0b11111111),  # Input Dir ALL
    ],
)
def test_pin_dir_in(reg_value, opcode, updated_reg_value):
    assert _apply_opcode(reg_value, opcode) == updated_reg_value



"""
OpCodes and Enums for thermocouple configuration and command operations
"""

from enum import Enum, unique
from edgepi.reg_helper.reg_helper import OpCode

REG_SIZE = 8  # MAX31856 register size in bits
NUM_WRITE_REGS = 12 # Number of MAX31856 writeable registers


@unique
class TCAddresses(Enum):
    """Valid hexadecimal address values for MAX31856 registers"""

    # Read addresses
    CR0_R = 0x00
    CR1_R = 0x01
    MASK_R = 0x02
    CJHF_R = 0x03
    CJLF_R = 0x04
    LTHFTH_R = 0x05
    LTHFTL_R = 0x06
    LTLFTH_R = 0x07
    LTLFTL_R = 0x08
    CJTO_R = 0x09
    CJTH_R = 0x0A
    CJTL_R = 0x0B
    LTCBH_R = 0x0C
    LTCBM_R = 0x0D
    LTCBL_R = 0x0E
    SR_R = 0x0F

    # Write addresses
    CR0_W = 0x80
    CR1_W = 0x81
    MASK_W = 0x82
    CJHF_W = 0x83
    CJLF_W = 0x84
    LTHFTH_W = 0x85
    LTHFTL_W = 0x86
    LTLFTH_W = 0x87
    LTLFTL_W = 0x88
    CJTO_W = 0x89
    CJTH_W = 0x8A
    CJTL_W = 0x8B


class Masks(Enum):
    """values for clearing or 'masking' register bits"""

    BIT0_MASK = 0xFE
    BIT1_MASK = 0xFD
    BIT2_MASK = 0xFB
    BIT3_MASK = 0xF7
    BIT4_MASK = 0xEF
    BIT5_MASK = 0xDF
    BIT6_MASK = 0xBF
    BIT7_MASK = 0x7F
    CR0_OC_MASK = 0xCF
    CR1_HIGH_MASK = 0x0F
    CR1_LOW_MASK = 0xF0
    BYTE_MASK = 0x00


@unique
class TempBits(Enum):
    """number of bits used in MAX31856 temperature registers to store values"""

    CJ_BITS = 14
    LT_BITS = 19
    CJ_DECIMAL_BITS = 6
    LT_DECIMAL_BITS = 7


@unique
class TCOps(Enum):
    """valid opcodes for commands that can be sent to thermocouple"""

    SINGLE_SHOT = OpCode(
        0x40, TCAddresses.CR0_W.value, Masks.BIT6_MASK.value
    )  # trigger a single temperature conversion
    CLEAR_FAULTS = OpCode(
        0x02, TCAddresses.CR0_W.value, Masks.BIT1_MASK.value
    )  # clear fault status register, only use with Interrupt Fault Mode


@unique
class DecBits4(Enum):
    """
    Valid decimal values for temperature registers with precision up to 2^-4.
    The MAX31856 uses a custom bit value scheme for setting temperature decimal values.
    One such scheme allocates 4 bits for this purpose, meaning only 16 decimal values are
    possible.

    These are listed below. Note, these values are not intended to be combined, as there
    is no guarantee the sum value can be represented. Please only select the exact value
    you need from this list.

    Examples:
        P0_5 represents the decimal value 0.5
        P0 represents the decimal value 0.0
        P0_0625 represents the decimal value 0.0625
    """

    P0 = 0b0000
    P0_5 = 0b1000
    P0_75 = 0b1100
    P0_875 = 0b1110
    P0_9375 = 0b1111
    P0_4375 = 0b0111
    P0_1875 = 0b0011
    P0_0625 = 0b0001
    P0_5625 = 0b1001
    P0_8125 = 0b1101
    P0_6875 = 0b1011
    P0_25 = 0b0100
    P0_125 = 0b0010
    P0_625 = 0b1010
    P0_3125 = 0b0101
    P0_375 = 0b0110


@unique
class DecBits6(Enum):
    """
    Valid decimal values for temperature registers with precision up to 2^-6.
    The MAX31856 uses a custom bit value scheme for setting temperature decimal values.
    One such scheme allocates 6 bits for this purpose, meaning only 64 decimal values are
    possible.

    These are listed below. Note, these values are not intended to be combined, as there
    is no guarantee the sum value can be represented. Please only select the exact value
    you need from this list.

    Examples:
        P0_5 represents the decimal value 0.5
        P0 represents the decimal value 0.0
        P0_0625 represents the decimal value 0.0625
    """

    P0 = 0b000000
    P0_015625 = 0b000001
    P0_03125 = 0b000010
    P0_046875 = 0b000011
    P0_0625 = 0b000100
    P0_078125 = 0b000101
    P0_09375 = 0b000110
    P0_109375 = 0b000111
    P0_125 = 0b001000
    P0_140625 = 0b001001
    P0_15625 = 0b001010
    P0_171875 = 0b001011
    P0_1875 = 0b001100
    P0_203125 = 0b001101
    P0_21875 = 0b001110
    P0_234375 = 0b001111
    P0_25 = 0b010000
    P0_265625 = 0b010001
    P0_28125 = 0b010010
    P0_296875 = 0b010011
    P0_3125 = 0b010100
    P0_328125 = 0b010101
    P0_34375 = 0b010110
    P0_359375 = 0b010111
    P0_375 = 0b011000
    P0_390625 = 0b011001
    P0_40625 = 0b011010
    P0_421875 = 0b011011
    P0_4375 = 0b011100
    P0_453125 = 0b011101
    P0_46875 = 0b011110
    P0_484375 = 0b011111
    P0_5 = 0b100000
    P0_515625 = 0b100001
    P0_53125 = 0b100010
    P0_546875 = 0b100011
    P0_5625 = 0b100100
    P0_578125 = 0b100101
    P0_59375 = 0b100110
    P0_609375 = 0b100111
    P0_625 = 0b101000
    P0_640625 = 0b101001
    P0_65625 = 0b101010
    P0_671875 = 0b101011
    P0_6875 = 0b101100
    P0_703125 = 0b101101
    P0_71875 = 0b101110
    P0_734375 = 0b101111
    P0_75 = 0b110000
    P0_765625 = 0b110001
    P0_78125 = 0b110010
    P0_796875 = 0b110011
    P0_8125 = 0b110100
    P0_828125 = 0b110101
    P0_84375 = 0b110110
    P0_859375 = 0b110111
    P0_875 = 0b111000
    P0_890625 = 0b111001
    P0_90625 = 0b111010
    P0_921875 = 0b111011
    P0_9375 = 0b111100
    P0_953125 = 0b111101
    P0_96875 = 0b111110
    P0_984375 = 0b111111


@unique
class ConvMode(Enum):
    """valid opcodes for setting thermocouple conversion mode"""

    SINGLE = OpCode(
        0x00, TCAddresses.CR0_W.value, Masks.BIT7_MASK.value
    )  # set thermocouple to perform single, manually triggered conversions
    AUTO = OpCode(
        0x80, TCAddresses.CR0_W.value, Masks.BIT7_MASK.value
    )  # set thermocouple to perform continuous conversions


@unique
class CJMode(Enum):
    """valid opcodes for setting thermocouple cold junction mode"""

    ENABLE = OpCode(
        0x00, TCAddresses.CR0_W.value, Masks.BIT3_MASK.value
    )  # enable the cold-junction sensor
    DISABLE = OpCode(
        0x08, TCAddresses.CR0_W.value, Masks.BIT3_MASK.value
    )  # disable the cold-junction sensor


@unique
class FaultMode(Enum):
    """valid opcodes for setting thermocouple fault mode"""

    COMPARATOR = OpCode(
        0x00, TCAddresses.CR0_W.value, Masks.BIT2_MASK.value
    )  # faults deassert only once fault condition is no longer true
    INTERRUPT = OpCode(
        0x04, TCAddresses.CR0_W.value, Masks.BIT2_MASK.value
    )  # faults deassert only once TCOps.CLEAR_FAULTS command is issued


@unique
class NoiseFilterMode(Enum):
    """valid opcodes for setting thermocouple noise rejection filter mode"""

    HZ_60 = OpCode(
        0x00, TCAddresses.CR0_W.value, Masks.BIT0_MASK.value
    )  # reject 60 Hz and its harmonics
    HZ_50 = OpCode(
        0x01, TCAddresses.CR0_W.value, Masks.BIT0_MASK.value
    )  # reject 50 Hz and its harmonics


@unique
class AvgMode(Enum):
    """valid opcodes for setting thermocouple conversion averaging mode"""

    AVG_1 = OpCode(0x00, TCAddresses.CR1_W.value, Masks.CR1_HIGH_MASK.value)  # single sample
    AVG_2 = OpCode(0x10, TCAddresses.CR1_W.value, Masks.CR1_HIGH_MASK.value)  # 2 samples averaged
    AVG_4 = OpCode(0x20, TCAddresses.CR1_W.value, Masks.CR1_HIGH_MASK.value)  # 4 samples averaged
    AVG_8 = OpCode(0x30, TCAddresses.CR1_W.value, Masks.CR1_HIGH_MASK.value)  # 8 samples averaged
    AVG_16 = OpCode(0x40, TCAddresses.CR1_W.value, Masks.CR1_HIGH_MASK.value)  # 16 samples averaged


@unique
class TCType(Enum):
    """valid opcodes for setting thermocouple type"""

    TYPE_B = OpCode(0x00, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value)  # type B thermocouple
    TYPE_E = OpCode(0x01, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value)  # type E thermocouple
    TYPE_J = OpCode(0x02, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value)  # type J thermocouple
    TYPE_K = OpCode(0x03, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value)  # type K thermocouple
    TYPE_N = OpCode(0x04, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value)  # type N thermocouple
    TYPE_R = OpCode(0x05, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value)  # type R thermocouple
    TYPE_S = OpCode(0x06, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value)  # type S thermocouple
    TYPE_T = OpCode(0x07, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value)  # type T thermocouple


class VoltageMode(Enum):
    """
    valid opcodes for setting thermocouple voltage mode. Use to set
    thermocouple type other than those listed under TCType.
    Note, When voltage mode is selected, no linearization is
    performed on the conversion data. Use the voltage data
    and the cold-junction temperature to calculate the thermo-
    couple’s hot-junction temperature.
    """

    GAIN_8 = OpCode(
        0x08, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value
    )  # full-scale input voltage range of +/- 78.125 mV
    GAIN_32 = OpCode(
        0x0C, TCAddresses.CR1_W.value, Masks.CR1_LOW_MASK.value
    )  # full-scale input voltage range of +/- 19.531 mV


# Below are valid opcodes for setting the thermocouple fault mask register
# Example: OPEN_MASK_ON will 'mask' the OPEN fault from asserting through
# the FAULT pin, but the fault register will still be updated. OPEN_MASK_OFF
# will allow the OPEN fault to assert through the FAULT pin. Note, the FAULT
# pin is currently not connected on the EdgePi, so these settings do not currently
# result in any changes to the thermocouple functionality.
# Note, the MAX31856 does not permit unmasking the TCRANGE and CJRANGE faults.


@unique
class OpenMask(Enum):
    """valid opcodes for setting the thermocouple OPEN fault mask"""

    OPEN_MASK_ON = OpCode(0x01, TCAddresses.MASK_W.value, Masks.BIT0_MASK.value)
    OPEN_MASK_OFF = OpCode(0x00, TCAddresses.MASK_W.value, Masks.BIT0_MASK.value)


@unique
class OvuvMask(Enum):
    """valid opcodes for setting the thermocouple OVUV fault mask"""

    OVUV_MASK_ON = OpCode(0x02, TCAddresses.MASK_W.value, Masks.BIT1_MASK.value)
    OVUV_MASK_OFF = OpCode(0x00, TCAddresses.MASK_W.value, Masks.BIT1_MASK.value)


@unique
class TCLowMask(Enum):
    """valid opcodes for setting the thermocouple TCLOW fault mask"""

    TCLOW_MASK_ON = OpCode(0x04, TCAddresses.MASK_W.value, Masks.BIT2_MASK.value)
    TCLOW_MASK_OFF = OpCode(0x00, TCAddresses.MASK_W.value, Masks.BIT2_MASK.value)


@unique
class TCHighMask(Enum):
    """valid opcodes for setting the thermocouple TCHIGH fault mask"""

    TCHIGH_MASK_ON = OpCode(0x08, TCAddresses.MASK_W.value, Masks.BIT3_MASK.value)
    TCHIGH_MASK_OFF = OpCode(0x00, TCAddresses.MASK_W.value, Masks.BIT3_MASK.value)


@unique
class CJLowMask(Enum):
    """valid opcodes for setting the thermocouple CJLOW fault mask"""

    CJLOW_MASK_ON = OpCode(0x10, TCAddresses.MASK_W.value, Masks.BIT4_MASK.value)
    CJLOW_MASK_OFF = OpCode(0x00, TCAddresses.MASK_W.value, Masks.BIT4_MASK.value)


@unique
class CJHighMask(Enum):
    """valid opcodes for setting the thermocouple CJHIGH fault mask"""

    CJHIGH_MASK_ON = OpCode(0x20, TCAddresses.MASK_W.value, Masks.BIT5_MASK.value)
    CJHIGH_MASK_OFF = OpCode(0x00, TCAddresses.MASK_W.value, Masks.BIT5_MASK.value)


@unique
class OpenCircuitMode(Enum):
    """valid opcodes for setting thermocouple open circuit fault detection mode"""

    DISABLED = OpCode(0x00, TCAddresses.CR0_W.value, Masks.CR0_OC_MASK.value)
    LOW_INPUT_IMPEDANCE = OpCode(0x10, TCAddresses.CR0_W.value, Masks.CR0_OC_MASK.value)
    MED_INPUT_IMPEDANCE = OpCode(0x20, TCAddresses.CR0_W.value, Masks.CR0_OC_MASK.value)
    HIGH_INPUT_IMPEDANCE = OpCode(0x30, TCAddresses.CR0_W.value, Masks.CR0_OC_MASK.value)



"""helper module to read serialized byte string"""
import os
PATH = os.path.dirname(os.path.abspath(__file__))

def read_binfile():
    """Read the dummy serializedFile and return byte string"""
    with open(PATH+"/serializedFile","rb") as fd:
        b_string = fd.read()
    return b_string



'''unit test for edgepi_calibration'''

# pylint: disable=C0413
# pylint: disable=protected-access

from unittest import mock
import os
PATH = os.path.dirname(os.path.abspath(__file__))
import sys
sys.modules['periphery'] = mock.MagicMock()

import pytest
from edgepi.calibration.edgepi_calibration import EdgePiCalibration
from edgepi.calibration.eeprom_constants import ModuleNames

def read_binfile():
    """Read the dummy serializedFile and return byte string"""
    with open(PATH+"/serializedFile","rb") as fd:
        b_string = fd.read()
    return b_string

# @pytest.fixture(name="calib")
# def fixture_test_dac():
#     yield EdgePiCalibration()

@pytest.mark.parametrize("module_name, result", [(ModuleNames.DAC, [0, 8]),
                                                 (ModuleNames.ADC, [1, 8]),
                                                 (ModuleNames.RTD, [2, 1]),
                                                 (ModuleNames.TC, [3, 1])])
def test_init_class(module_name, result):
    edge_calib = EdgePiCalibration(module_name)
    assert edge_calib.module == result[0]
    assert edge_calib.num_of_ch == result[1]

@pytest.mark.parametrize("module_name, num_of_points, result", [(ModuleNames.DAC, 10, 0),
                                                                (ModuleNames.ADC, 5, 0),
                                                                (ModuleNames.RTD, 2, 0),
                                                                (ModuleNames.TC,  7, 0)])
def test_generate_measurements_dict(module_name, num_of_points, result):
    edge_calib = EdgePiCalibration(module_name)
    measurements_dict = edge_calib.generate_measurements_dict(num_of_points)
    for _, value in measurements_dict.items():
        assert value['input_unit'] == result
        assert value['expected_out'] == result
        assert value['actual_out'] == result
    dict_length = len(measurements_dict.keys())
    assert dict_length == num_of_points

@pytest.mark.parametrize("module_name, num_of_points, result", [(ModuleNames.DAC, 10, 0),
                                                                (ModuleNames.ADC, 5, 0),
                                                                (ModuleNames.RTD, 2, 0),
                                                                (ModuleNames.TC,  7, 0)])
def test_generate_channel_measurements_dict(module_name, num_of_points, result):
    edge_calib = EdgePiCalibration(module_name)
    measurements_dict = edge_calib.generate_channel_measurements_dict(num_of_points)
    for ch in range(edge_calib.num_of_ch):
        for keys, _ in measurements_dict[ch].items():
            assert measurements_dict[ch][keys]['input_unit'] == result
            assert measurements_dict[ch][keys]['expected_out'] == result
            assert measurements_dict[ch][keys]['actual_out'] == result
    dict_length = len(measurements_dict.keys())
    assert dict_length == edge_calib.num_of_ch

@pytest.mark.parametrize("module_name, num_of_points, values_to_record",
                        [(ModuleNames.DAC, 10, [[2, 3.45, 5.34],
                                                [26, 3245.24, 123.13],
                                                [25, 325.224, 123.123],
                                                [56, 345.224, 12.123],
                                                [456, 3245.24, 123.123],
                                                [2456, 325.224, 123.23],
                                                [246, 3245.224, 123.123],
                                                [256, 3245.24, 123.13],
                                                [246, 325.224, 123.13],
                                                [245, 325.24, 13.123]]),
                         (ModuleNames.ADC, 5, [[2, 3.45, 5.34],
                                               [246, 325.4, 123.1],
                                               [246, 325.24, 123.3],
                                               [246, 325.22, 12.13],
                                               [246, 3.224, 1.13]]),
                         (ModuleNames.RTD, 2, [[2, 3.45, 5.34],
                                               [246, 325.224, 123.13]]),
                         (ModuleNames.TC,  7, [[2, 3.45, 5.34],
                                              [246, 325.2, 123.13],
                                              [246, 325.224, 23.13],
                                              [246, 325.224, 13.13],
                                              [246, 325.224, 12.13],
                                              [246, 325.224, 123.3],
                                              [246, 325.224, 123.1]])
                        ])
def test_record_measurements(module_name, num_of_points, values_to_record):
    edge_calib = EdgePiCalibration(module_name)
    measurements_dict = edge_calib.generate_measurements_dict(num_of_points)
    for key, value_dict in measurements_dict.items():
        edge_calib.record_measurements(value_dict, values_to_record[key][0],
                                                   values_to_record[key][1],
                                                   values_to_record[key][2])
        assert value_dict['input_unit'] != 0
        assert value_dict['expected_out'] != 0
        assert value_dict['actual_out'] != 0



"""unit tests for edgepi_gpio module"""

# pylint: disable=C0413

from unittest import mock
import sys
if sys.platform != 'linux':
    sys.modules['periphery'] = mock.MagicMock()

from contextlib import nullcontext as does_not_raise
import pytest
from edgepi.gpio.gpio_constants import GpioPins
from edgepi.digital_input.edgepi_digital_input import EdgePiDigitalInput, InvalidPinName

@pytest.fixture(name="din")
def fixture_test_dac(mocker):
    mocker.patch('edgepi.gpio.edgepi_gpio.EdgePiGPIO')
    yield EdgePiDigitalInput()

@pytest.mark.parametrize("pin_name, mock_value, result, error",
                        [(GpioPins.DIN1, True, True, does_not_raise()),
                         (GpioPins.DIN2, True, True, does_not_raise()),
                         (GpioPins.DIN3, True, True, does_not_raise()),
                         (GpioPins.DIN4, True, True, does_not_raise()),
                         (GpioPins.DIN5, True, True, does_not_raise()),
                         (GpioPins.DIN6, True, True, does_not_raise()),
                         (GpioPins.DIN7, True, True, does_not_raise()),
                         (GpioPins.DIN8, True, True, does_not_raise()),
                         (GpioPins.DIN1, False, False, does_not_raise()),
                         (GpioPins.DIN2, False, False, does_not_raise()),
                         (GpioPins.DIN3, False, False, does_not_raise()),
                         (GpioPins.DIN4, False, False, does_not_raise()),
                         (GpioPins.DIN5, False, False, does_not_raise()),
                         (GpioPins.DIN6, False, False, does_not_raise()),
                         (GpioPins.DIN7, False, False, does_not_raise()),
                         (GpioPins.DIN8, False, False, does_not_raise()),
                         (GpioPins.DOUT2, False, False, pytest.raises(InvalidPinName)),
                         (None, False, False, pytest.raises(InvalidPinName))])
def test_edgepi_digital_input_state(mocker, pin_name, mock_value, result, error, din):
    mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIOChip.read_gpio_pin_state",
                 return_value = mock_value)
    with error:
        state = din.digital_input_state(pin_name)
        assert state == result
    


'''unit test for access eeprom'''

# pylint: disable=C0413


from unittest import mock
import os
PATH = os.path.dirname(os.path.abspath(__file__))
import sys
sys.modules['periphery'] = mock.MagicMock()

import pytest
from edgepi.calibration.eeprom_constants import MessageFieldNumber
from edgepi.calibration.edgepi_eeprom import EdgePiEEPROM
from edgepi.calibration.eeprom_mapping_pb2 import EepromLayout

@pytest.fixture(name="eeprom")
def fixture_test_dac(mocker):
    mocker.patch("edgepi.peripherals.i2c.I2C")
    yield EdgePiEEPROM()

def read_binfile():
    """Read the dummy serializedFile and return byte string"""
    with open(PATH+"/serializedFile","rb") as fd:
        b_string = fd.read()
    return b_string

@pytest.mark.parametrize("page_addr, byte_addr, result",
                        [(0, 8, [0x00, 0x08]),
                         (0, 63, [0x00, 0x3F]),
                         (511, 63, [0x7F, 0xFF]),
                         (511, 0, [0x7F, 0xC0])
                        ])
def test__pack_mem_address(page_addr, byte_addr, result, eeprom):
    # pylint: disable=protected-access
    address_message = eeprom._EdgePiEEPROM__pack_mem_address(page_addr, byte_addr)
    assert address_message == result

@pytest.mark.parametrize("memory_address, result",
                        [(0, [0, 0]),
                         (63, [0, 63]),
                         (64, [1, 0]),
                         (32767, [511, 63]),
                         (32704, [511, 0])
                        ])
def test__byte_address_generation(memory_address, result, eeprom):
    # pylint: disable=protected-access
    page_addr, byte_addr = eeprom._EdgePiEEPROM__byte_address_generation(memory_address)
    assert page_addr == result[0]
    assert byte_addr == result[1]

@pytest.mark.parametrize("reg_addr, mock_val, result", [(1, [23], [23])])
def test_selective_read(mocker, eeprom, reg_addr, mock_val, result):
    mocker.patch("edgepi.peripherals.i2c.I2CDevice.transfer",return_value = mock_val)
    read_result = eeprom.selective_read(reg_addr)
    assert read_result == result



@pytest.mark.parametrize("reg_addr, length, mock_val, result", [(1, 5,
                                                                [23, 34, 56, 7, 8],
                                                                [23, 34, 56, 7, 8])])
def test_sequential_read(mocker, eeprom, reg_addr, length, mock_val, result):
    mocker.patch("edgepi.peripherals.i2c.I2CDevice.transfer",return_value = mock_val)
    read_result = eeprom.sequential_read( reg_addr, length)
    assert read_result == result

@pytest.mark.parametrize("mock_value,result",
                        [([0,123], 123),
                         ([1,1], 257),
                         ([63,255], 16383),
                         ([63, 0], 16128)
                        ])
def test__allocated_memory(mocker,mock_value,result, eeprom):
    # pylint: disable=protected-access
    mocker.patch("edgepi.calibration.edgepi_eeprom.EdgePiEEPROM.sequential_read",
                return_value =mock_value)
    length = eeprom._EdgePiEEPROM__allocated_memory()
    assert length == result

def test__read_edgepi_reserved_memory(mocker, eeprom):
    # pylint: disable=protected-access
    mocker.patch("edgepi.calibration.edgepi_eeprom.EdgePiEEPROM._EdgePiEEPROM__allocated_memory")
    mocker.patch("edgepi.calibration.edgepi_eeprom.EdgePiEEPROM.sequential_read",
                return_value =list(read_binfile()))
    byte_string = eeprom._EdgePiEEPROM__read_edgepi_reserved_memory()
    assert byte_string == read_binfile()

@pytest.mark.parametrize("msg",
                        [(MessageFieldNumber.DAC),
                         (MessageFieldNumber.ADC),
                         (MessageFieldNumber.RTD),
                         (MessageFieldNumber.TC),
                         (MessageFieldNumber.CONFIGS_KEY),
                         (MessageFieldNumber.DATA_KEY),
                         (MessageFieldNumber.SERIAL),
                         (MessageFieldNumber.MODEL),
                         (MessageFieldNumber.CLIENT_ID)
                        ])
def test_get_message_of_interest(mocker, msg, eeprom):
    # pylint: disable=protected-access
    mocker.patch(
        "edgepi.calibration.edgepi_eeprom.EdgePiEEPROM._EdgePiEEPROM__read_edgepi_reserved_memory",
        return_value = read_binfile())
    memory_contents = EepromLayout()
    memory_contents.ParseFromString(read_binfile())
    msg_of_interest = eeprom.get_message_of_interest(msg)
    assert msg_of_interest == memory_contents.ListFields()[msg.value -1][1]

def test_get_edgepi_reserved_data(mocker, eeprom):
    # pylint: disable=protected-access
    mocker.patch(
        "edgepi.calibration.edgepi_eeprom.EdgePiEEPROM._EdgePiEEPROM__read_edgepi_reserved_memory",
        return_value = read_binfile())
    memory_contents = EepromLayout()
    memory_contents.ParseFromString(read_binfile())
    edgepi_eeprom_dataclass = eeprom.get_edgepi_reserved_data()
    assert edgepi_eeprom_dataclass.dac_calib_parms is not None
    assert edgepi_eeprom_dataclass.adc_calib_parms is not None
    assert edgepi_eeprom_dataclass.rtd_calib_parms is not None
    assert edgepi_eeprom_dataclass.tc_calib_parms is not None
    assert edgepi_eeprom_dataclass.config_key is not None
    assert edgepi_eeprom_dataclass.data_key is not None
    assert edgepi_eeprom_dataclass.serial is not None
    assert edgepi_eeprom_dataclass.model is not None
    assert edgepi_eeprom_dataclass.client_id is not None



"""Digital Input Module"""

from edgepi.gpio.gpio_constants import GpioPins
from edgepi.gpio.gpio_configs import DINPins
from edgepi.gpio.edgepi_gpio import EdgePiGPIO

class InvalidPinName(Exception):
    """Raised invalid pin name"""

class EdgePiDigitalInput():
    """handling reading of digital inputs"""
    def __init__(self):
        # To limit access to input functionality, using composition rather than inheritance
        self.gpio = EdgePiGPIO()

    def digital_input_state(self, pin_name: GpioPins = None):
        """
        Read selected GPIO pin
        Args:
            pin_name (GpioPins): GpioPin enums
        Return:
            state (bool): corresponding pin state
        """
        if pin_name is None or pin_name.value not in [pins.value for pins in DINPins]:
            raise InvalidPinName(f'Invalid pin name passed: {pin_name}')
        return self.gpio.read_pin_state(pin_name.value)



"""
Module for SPI devices

Classes:
    SpiDevice
"""


import logging

from periphery import SPI


_logger = logging.getLogger(__name__)

# TODO: This class needs to be changed as the SPI library changes
class SpiDevice:
    """Class representing an I2C device"""

    _devPath = "/dev/spidev"

    def __init__(
        self,
        bus_num: int = None,
        dev_id: int = None,
        mode: int = 1,
        max_speed: int = 1000000,
        bit_order: str = "msb",
        bits_per_word: int = 8,
        extra_flags: int = 0,
    ):
        self.devpath = f"/dev/spidev{bus_num}.{dev_id}"
        _logger.debug(f"Initialized SPI device with path '{self.devpath}'")
        self.spi = SPI(
            self.devpath,
            mode,
            max_speed,
            bit_order,
            bits_per_word,
            extra_flags,
        )

    def transfer(self, data: list) -> list:
        """Conduct an SPI data transfer"""
        _logger.debug(f"Before SPI transfer: data={data}")
        out = self.spi.transfer(data)
        _logger.debug(f"After SPI transfer: data={out}")
        return out

    def close(self):
        """Close SPI connection"""
        self.spi.close()



"""
For ADC state querying
"""

from dataclasses import dataclass
from typing import Any
from enum import Enum

from edgepi.adc.adc_constants import (
    IDACMAG,
    IDACMUX,
    REFMUX,
    ADC1DataRate,
    ADC2DataRate,
    ADCChannel,
    ADCMasks,
    ADCReg,
    CheckMode,
    ConvMode,
    FilterMode,
    StatusByte,
)
from edgepi.reg_helper.reg_helper import BitMask


@dataclass
class PropertyValue:
    """
    Stores information about a specific value of an ADC functional adc_property

    Attributes
        `value` (any): user-friendly description of this adc_property value
        `code` (Enum): internal code used to set this adc_property value
    """

    value: Any
    code: Enum

    def __repr__(self) -> str:
        return f"value: {self.value}, code: {self.code}"


@dataclass
class Property:
    """Stores information about an ADC functional adc_property"""

    name: str
    addx: int
    mask: int
    values: dict[int, PropertyValue]


class ADCProperties(Enum):
    """Address and mask values for ADC properties"""

    CONV_MODE = Property(
        "conversion-adc_property",
        ADCReg.REG_MODE0.value,
        BitMask.BIT6.value,
        {
            ConvMode.PULSE.value.op_code: PropertyValue("pulse", ConvMode.PULSE),
            ConvMode.CONTINUOUS.value.op_code: PropertyValue("continuous", ConvMode.CONTINUOUS),
        },
    )
    CHECK_MODE = Property(
        "checksum-byte-adc_property",
        ADCReg.REG_INTERFACE.value,
        ADCMasks.CHECK_BITS.value,
        {
            CheckMode.CHECK_BYTE_OFF.value.op_code: PropertyValue("off", CheckMode.CHECK_BYTE_OFF),
            CheckMode.CHECK_BYTE_CHK.value.op_code: PropertyValue(
                "checksum", CheckMode.CHECK_BYTE_CHK
            ),
            CheckMode.CHECK_BYTE_CRC.value.op_code: PropertyValue("crc", CheckMode.CHECK_BYTE_CRC),
        },
    )
    STATUS_MODE = Property(
        "status-byte-adc_property",
        ADCReg.REG_INTERFACE.value,
        BitMask.BIT2.value,
        {
            StatusByte.STATUS_BYTE_OFF.value.op_code: PropertyValue(
                "off", StatusByte.STATUS_BYTE_OFF
            ),
            StatusByte.STATUS_BYTE_ON.value.op_code: PropertyValue("on", StatusByte.STATUS_BYTE_ON),
        },
    )
    DATA_RATE_1 = Property(
        "data-rate-adc-1",
        ADCReg.REG_MODE2.value,
        BitMask.LOW_NIBBLE.value,
        {
            ADC1DataRate.SPS_2P5.value.op_code: PropertyValue(2.5, ADC1DataRate.SPS_2P5),
            ADC1DataRate.SPS_5.value.op_code: PropertyValue(5, ADC1DataRate.SPS_5),
            ADC1DataRate.SPS_10.value.op_code: PropertyValue(10, ADC1DataRate.SPS_10),
            ADC1DataRate.SPS_16P6.value.op_code: PropertyValue(16.6, ADC1DataRate.SPS_16P6),
            ADC1DataRate.SPS_20.value.op_code: PropertyValue(20, ADC1DataRate.SPS_20),
            ADC1DataRate.SPS_50.value.op_code: PropertyValue(50, ADC1DataRate.SPS_50),
            ADC1DataRate.SPS_60.value.op_code: PropertyValue(60, ADC1DataRate.SPS_60),
            ADC1DataRate.SPS_100.value.op_code: PropertyValue(100, ADC1DataRate.SPS_100),
            ADC1DataRate.SPS_400.value.op_code: PropertyValue(400, ADC1DataRate.SPS_400),
            ADC1DataRate.SPS_1200.value.op_code: PropertyValue(1200, ADC1DataRate.SPS_1200),
            ADC1DataRate.SPS_2400.value.op_code: PropertyValue(2400, ADC1DataRate.SPS_2400),
            ADC1DataRate.SPS_4800.value.op_code: PropertyValue(4800, ADC1DataRate.SPS_4800),
            ADC1DataRate.SPS_7200.value.op_code: PropertyValue(7200, ADC1DataRate.SPS_7200),
            ADC1DataRate.SPS_14400.value.op_code: PropertyValue(14400, ADC1DataRate.SPS_14400),
            ADC1DataRate.SPS_19200.value.op_code: PropertyValue(19200, ADC1DataRate.SPS_19200),
            ADC1DataRate.SPS_38400.value.op_code: PropertyValue(38400, ADC1DataRate.SPS_38400),
        },
    )
    DATA_RATE_2 = Property(
        "data-rate-adc-2",
        ADCReg.REG_ADC2CFG.value,
        ADCMasks.ADC2_DR_BITS.value,
        {
            ADC2DataRate.SPS_10.value.op_code: PropertyValue(10, ADC2DataRate.SPS_10),
            ADC2DataRate.SPS_100.value.op_code: PropertyValue(10, ADC2DataRate.SPS_100),
            ADC2DataRate.SPS_400.value.op_code: PropertyValue(400, ADC2DataRate.SPS_400),
            ADC2DataRate.SPS_800.value.op_code: PropertyValue(800, ADC2DataRate.SPS_800),
        },
    )
    FILTER_MODE = Property(
        "filter-adc_property",
        ADCReg.REG_MODE1.value,
        ADCMasks.FILTER_BITS.value,
        {
            FilterMode.SINC1.value.op_code: PropertyValue("sinc-1", FilterMode.SINC1),
            FilterMode.SINC2.value.op_code: PropertyValue("sinc-2", FilterMode.SINC2),
            FilterMode.SINC3.value.op_code: PropertyValue("sinc-3", FilterMode.SINC3),
            FilterMode.SINC4.value.op_code: PropertyValue("sinc-4", FilterMode.SINC4),
            FilterMode.FIR.value.op_code: PropertyValue("FIR", FilterMode.FIR),
        },
    )
    ADC1_MUXP = Property(
        "adc-1-mux-p",
        ADCReg.REG_INPMUX.value,
        BitMask.HIGH_NIBBLE.value,
        {
            0x00: PropertyValue(0, ADCChannel.AIN0),
            0x10: PropertyValue(1, ADCChannel.AIN1),
            0x20: PropertyValue(2, ADCChannel.AIN2),
            0x30: PropertyValue(3, ADCChannel.AIN3),
            0x40: PropertyValue(4, ADCChannel.AIN4),
            0x50: PropertyValue(5, ADCChannel.AIN5),
            0x60: PropertyValue(6, ADCChannel.AIN6),
            0x70: PropertyValue(7, ADCChannel.AIN7),
            0x80: PropertyValue(8, ADCChannel.AIN8),
            0x90: PropertyValue(9, ADCChannel.AIN9),
            0xA0: PropertyValue("AINCOM", ADCChannel.AINCOM),
            0xF0: PropertyValue("floating", ADCChannel.FLOAT),
        },
    )
    ADC2_MUXP = Property(
        "adc-2-mux-p",
        ADCReg.REG_ADC2MUX.value,
        BitMask.HIGH_NIBBLE.value,
        {
            0x00: PropertyValue(0, ADCChannel.AIN0),
            0x10: PropertyValue(1, ADCChannel.AIN1),
            0x20: PropertyValue(2, ADCChannel.AIN2),
            0x30: PropertyValue(3, ADCChannel.AIN3),
            0x40: PropertyValue(4, ADCChannel.AIN4),
            0x50: PropertyValue(5, ADCChannel.AIN5),
            0x60: PropertyValue(6, ADCChannel.AIN6),
            0x70: PropertyValue(7, ADCChannel.AIN7),
            0x80: PropertyValue(8, ADCChannel.AIN8),
            0x90: PropertyValue(9, ADCChannel.AIN9),
            0xA0: PropertyValue("AINCOM", ADCChannel.AINCOM),
            0xF0: PropertyValue("floating", ADCChannel.FLOAT),
        },
    )
    ADC1_MUXN = Property(
        "adc-1-mux-n",
        ADCReg.REG_INPMUX.value,
        BitMask.LOW_NIBBLE.value,
        {
            0x0: PropertyValue(0, ADCChannel.AIN0),
            0x1: PropertyValue(1, ADCChannel.AIN1),
            0x2: PropertyValue(2, ADCChannel.AIN2),
            0x3: PropertyValue(3, ADCChannel.AIN3),
            0x4: PropertyValue(4, ADCChannel.AIN4),
            0x5: PropertyValue(5, ADCChannel.AIN5),
            0x6: PropertyValue(6, ADCChannel.AIN6),
            0x7: PropertyValue(7, ADCChannel.AIN7),
            0x8: PropertyValue(8, ADCChannel.AIN8),
            0x9: PropertyValue(9, ADCChannel.AIN9),
            0xA: PropertyValue("AINCOM", ADCChannel.AINCOM),
            0xF: PropertyValue("floating", ADCChannel.FLOAT),
        },
    )
    ADC2_MUXN = Property(
        "adc-2-mux-n",
        ADCReg.REG_ADC2MUX.value,
        BitMask.LOW_NIBBLE.value,
        {
            0x0: PropertyValue(0, ADCChannel.AIN0),
            0x1: PropertyValue(1, ADCChannel.AIN1),
            0x2: PropertyValue(2, ADCChannel.AIN2),
            0x3: PropertyValue(3, ADCChannel.AIN3),
            0x4: PropertyValue(4, ADCChannel.AIN4),
            0x5: PropertyValue(5, ADCChannel.AIN5),
            0x6: PropertyValue(6, ADCChannel.AIN6),
            0x7: PropertyValue(7, ADCChannel.AIN7),
            0x8: PropertyValue(8, ADCChannel.AIN8),
            0x9: PropertyValue(9, ADCChannel.AIN9),
            0xA: PropertyValue("AINCOM", ADCChannel.AINCOM),
            0xF: PropertyValue("floating", ADCChannel.FLOAT),
        },
    )
    IDAC1_MUX = Property(
        "idac-1-mux",
        ADCReg.REG_IDACMUX.value,
        BitMask.LOW_NIBBLE.value,
        {
            IDACMUX.IDAC1_AIN0.value.op_code: PropertyValue(0, IDACMUX.IDAC1_AIN0),
            IDACMUX.IDAC1_AIN1.value.op_code: PropertyValue(1, IDACMUX.IDAC1_AIN1),
            IDACMUX.IDAC1_AIN2.value.op_code: PropertyValue(2, IDACMUX.IDAC1_AIN2),
            IDACMUX.IDAC1_AIN3.value.op_code: PropertyValue(3, IDACMUX.IDAC1_AIN3),
            IDACMUX.IDAC1_AIN4.value.op_code: PropertyValue(4, IDACMUX.IDAC1_AIN4),
            IDACMUX.IDAC1_AIN5.value.op_code: PropertyValue(5, IDACMUX.IDAC1_AIN5),
            IDACMUX.IDAC1_AIN6.value.op_code: PropertyValue(6, IDACMUX.IDAC1_AIN6),
            IDACMUX.IDAC1_AIN7.value.op_code: PropertyValue(7, IDACMUX.IDAC1_AIN7),
            IDACMUX.IDAC1_AIN8.value.op_code: PropertyValue(8, IDACMUX.IDAC1_AIN8),
            IDACMUX.IDAC1_AIN9.value.op_code: PropertyValue(9, IDACMUX.IDAC1_AIN9),
            IDACMUX.IDAC1_AINCOM.value.op_code: PropertyValue("AINCOM", IDACMUX.IDAC1_AINCOM),
            IDACMUX.IDAC1_NO_CONNECT.value.op_code: PropertyValue(
                "no-connection", IDACMUX.IDAC1_NO_CONNECT
            ),
        },
    )
    IDAC2_MUX = Property(
        "idac-2-mux",
        ADCReg.REG_IDACMUX.value,
        BitMask.HIGH_NIBBLE.value,
        {
            IDACMUX.IDAC2_AIN0.value.op_code: PropertyValue(0, IDACMUX.IDAC2_AIN0),
            IDACMUX.IDAC2_AIN1.value.op_code: PropertyValue(1, IDACMUX.IDAC2_AIN1),
            IDACMUX.IDAC2_AIN2.value.op_code: PropertyValue(2, IDACMUX.IDAC2_AIN2),
            IDACMUX.IDAC2_AIN3.value.op_code: PropertyValue(3, IDACMUX.IDAC2_AIN3),
            IDACMUX.IDAC2_AIN4.value.op_code: PropertyValue(4, IDACMUX.IDAC2_AIN4),
            IDACMUX.IDAC2_AIN5.value.op_code: PropertyValue(5, IDACMUX.IDAC2_AIN5),
            IDACMUX.IDAC2_AIN6.value.op_code: PropertyValue(6, IDACMUX.IDAC2_AIN6),
            IDACMUX.IDAC2_AIN7.value.op_code: PropertyValue(7, IDACMUX.IDAC2_AIN7),
            IDACMUX.IDAC2_AIN8.value.op_code: PropertyValue(8, IDACMUX.IDAC2_AIN8),
            IDACMUX.IDAC2_AIN9.value.op_code: PropertyValue(9, IDACMUX.IDAC2_AIN9),
            IDACMUX.IDAC2_AINCOM.value.op_code: PropertyValue("AINCOM", IDACMUX.IDAC2_AINCOM),
            IDACMUX.IDAC2_NO_CONNECT.value.op_code: PropertyValue(
                "no-connection", IDACMUX.IDAC2_NO_CONNECT
            ),
        },
    )
    IDAC1_MAG = Property(
        "idac-1-mag",
        ADCReg.REG_IDACMAG.value,
        BitMask.LOW_NIBBLE.value,
        {
            IDACMAG.IDAC1_OFF.value.op_code: PropertyValue(0, IDACMAG.IDAC1_OFF),
            IDACMAG.IDAC1_50.value.op_code: PropertyValue(50, IDACMAG.IDAC1_50),
            IDACMAG.IDAC1_100.value.op_code: PropertyValue(100, IDACMAG.IDAC1_100),
            IDACMAG.IDAC1_250.value.op_code: PropertyValue(250, IDACMAG.IDAC1_250),
            IDACMAG.IDAC1_500.value.op_code: PropertyValue(500, IDACMAG.IDAC1_500),
            IDACMAG.IDAC1_750.value.op_code: PropertyValue(750, IDACMAG.IDAC1_750),
            IDACMAG.IDAC1_1000.value.op_code: PropertyValue(1000, IDACMAG.IDAC1_1000),
            IDACMAG.IDAC1_1500.value.op_code: PropertyValue(1500, IDACMAG.IDAC1_1500),
            IDACMAG.IDAC1_2000.value.op_code: PropertyValue(2000, IDACMAG.IDAC1_2000),
            IDACMAG.IDAC1_2500.value.op_code: PropertyValue(2500, IDACMAG.IDAC1_2500),
            IDACMAG.IDAC1_3000.value.op_code: PropertyValue(3000, IDACMAG.IDAC1_3000),
        },
    )
    IDAC2_MAG = Property(
        "idac-2-mag",
        ADCReg.REG_IDACMAG.value,
        BitMask.HIGH_NIBBLE.value,
        {
            IDACMAG.IDAC2_OFF.value.op_code: PropertyValue(0, IDACMAG.IDAC2_OFF),
            IDACMAG.IDAC2_50.value.op_code: PropertyValue(50, IDACMAG.IDAC2_50),
            IDACMAG.IDAC2_100.value.op_code: PropertyValue(100, IDACMAG.IDAC2_100),
            IDACMAG.IDAC2_250.value.op_code: PropertyValue(250, IDACMAG.IDAC2_250),
            IDACMAG.IDAC2_500.value.op_code: PropertyValue(500, IDACMAG.IDAC2_500),
            IDACMAG.IDAC2_750.value.op_code: PropertyValue(750, IDACMAG.IDAC2_750),
            IDACMAG.IDAC2_1000.value.op_code: PropertyValue(1000, IDACMAG.IDAC2_1000),
            IDACMAG.IDAC2_1500.value.op_code: PropertyValue(1500, IDACMAG.IDAC2_1500),
            IDACMAG.IDAC2_2000.value.op_code: PropertyValue(2000, IDACMAG.IDAC2_2000),
            IDACMAG.IDAC2_2500.value.op_code: PropertyValue(2500, IDACMAG.IDAC2_2500),
            IDACMAG.IDAC2_3000.value.op_code: PropertyValue(3000, IDACMAG.IDAC2_3000),
        },
    )
    REFMUX_POS = Property(
        "ref-mux-positive",
        ADCReg.REG_REFMUX.value,
        ADCMasks.RMUXP_BITS.value,
        {
            REFMUX.POS_REF_INT_2P5.value.op_code: PropertyValue(
                "2.5 V internal", REFMUX.POS_REF_INT_2P5
            ),
            REFMUX.POS_REF_EXT_AIN0.value.op_code: PropertyValue(
                "channel 0 external", REFMUX.POS_REF_EXT_AIN0
            ),
            REFMUX.POS_REF_EXT_AIN2.value.op_code: PropertyValue(
                "channel 2 external", REFMUX.POS_REF_EXT_AIN2
            ),
            REFMUX.POS_REF_EXT_AIN4.value.op_code: PropertyValue(
                "channel 4 external", REFMUX.POS_REF_EXT_AIN4
            ),
            REFMUX.POS_REF_INT_VAVDD.value.op_code: PropertyValue(
                "VAVDD", REFMUX.POS_REF_INT_VAVDD
            ),
        },
    )
    REFMUX_NEG = Property(
        "ref-mux-negative",
        ADCReg.REG_REFMUX.value,
        ADCMasks.RMUXN_BITS.value,
        {
            REFMUX.NEG_REF_INT_2P5.value.op_code: PropertyValue(
                "2.5 V internal", REFMUX.NEG_REF_INT_2P5
            ),
            REFMUX.NEG_REF_EXT_AIN1.value.op_code: PropertyValue(
                "channel 1 external", REFMUX.NEG_REF_EXT_AIN1
            ),
            REFMUX.NEG_REF_EXT_AIN3.value.op_code: PropertyValue(
                "channel 3 external", REFMUX.NEG_REF_EXT_AIN3
            ),
            REFMUX.NEG_REF_EXT_AIN5.value.op_code: PropertyValue(
                "channel 5 external", REFMUX.NEG_REF_EXT_AIN5
            ),
            REFMUX.NEG_REF_INT_VAVSS.value.op_code: PropertyValue(
                "VAVSS", REFMUX.NEG_REF_INT_VAVSS
            ),
        },
    )



"""unit test for utilities.py module"""

import pytest
from bitstring import pack
from edgepi.utilities.utilities import filter_dict, bitstring_from_list
from edgepi.tc.tc_constants import AvgMode


@pytest.mark.parametrize(
    "args_dict, key, val, out",
    [
        ({"self": 0x0}, "self", "", {}),
        ({"avg_mode": AvgMode.AVG_1}, "self", "", {"avg_mode": AvgMode.AVG_1}),
        (
            {"self": 0x0, "avg_mode": AvgMode.AVG_1},
            "self",
            "",
            {"avg_mode": AvgMode.AVG_1},
        ),
        ({"self": 0x0, "avg_mode": AvgMode.AVG_1}, "avg_mode", "", {"self": 0x0}),
        ({"self": 0x0, "avg_mode": AvgMode.AVG_1}, "", AvgMode.AVG_1, {"self": 0x0}),
        (
            {"self": 0x0, "avg_mode": AvgMode.AVG_1},
            "",
            0x0,
            {"avg_mode": AvgMode.AVG_1},
        ),
        (
            {"self": 0x0, "avg_mode": AvgMode.AVG_1, "test_var": 0x5},
            "test_var",
            0x0,
            {"avg_mode": AvgMode.AVG_1},
        ),
        (
            {"self": 0x0, "avg_mode": AvgMode.AVG_1, "test_var": 0x5},
            "test_var",
            0x5,
            {"self": 0x0, "avg_mode": AvgMode.AVG_1},
        ),
    ],
)
def test_filter_dict(args_dict, key, val, out):
    assert filter_dict(args_dict, key, val) == out


@pytest.mark.parametrize("data_bytes, expected", [
    ([0x0, 0x0, 0x0], pack("uint:24", 0)),
    ([0x0, 0x0, 0x0, 0x0], pack("uint:32", 0)),
    ([0x01, 0x02, 0x03, 0x04], pack("hex:32", "0x01020304")),
    ([0x1, 0x2, 0x3, 0x4], pack("hex:32", "0x01020304")),
    ([0x10, 0x20, 0x30, 0x40], pack("hex:32", "0x10203040")),
    ([0xFF, 0xFF, 0xFF, 0xFF], pack("hex:32", "0xFFFFFFFF")),
])
def test_bitstring_from_list(data_bytes, expected):
    assert bitstring_from_list(data_bytes) == expected



"""" Unit tests for adc_multiplexers.py """


from contextlib import nullcontext as does_not_raise

import pytest
from edgepi.adc.adc_constants import ADCChannel as CH, ADCReg
from edgepi.reg_helper.reg_helper import BitMask, OpCode
from edgepi.adc.adc_multiplexers import (
    generate_mux_opcodes,
    ChannelNotAvailableError,
    validate_channels_allowed,
)


@pytest.mark.parametrize(
    "mux_updates, expected",
    [
        (
            {
                ADCReg.REG_INPMUX: (None, None),
                ADCReg.REG_ADC2MUX: (None, None),
            },
            [],
        ),
        (
            {
                ADCReg.REG_INPMUX: (None, CH.AIN2),
                ADCReg.REG_ADC2MUX: (None, None),
            },
            [],
        ),
        (
            {
                ADCReg.REG_INPMUX: (CH.AIN1, CH.AINCOM),
                ADCReg.REG_ADC2MUX: (None, None),
            },
            [OpCode(0x1A, ADCReg.REG_INPMUX.value, BitMask.BYTE.value)],
        ),
        (
            {
                ADCReg.REG_INPMUX: (CH.AIN7, None),
                ADCReg.REG_ADC2MUX: (None, None),
            },
            [],
        ),
        (
            {
                ADCReg.REG_INPMUX: (None, CH.AIN5),
                ADCReg.REG_ADC2MUX: (None, None),
            },
            [],
        ),
        (
            {
                ADCReg.REG_INPMUX: (None, None),
                ADCReg.REG_ADC2MUX: (CH.AIN5, None),
            },
            [],
        ),
        (
            {
                ADCReg.REG_INPMUX: (None, None),
                ADCReg.REG_ADC2MUX: (None, CH.AIN6),
            },
            [],
        ),
        (
            {
                ADCReg.REG_INPMUX: (None, None),
                ADCReg.REG_ADC2MUX: (CH.AIN5, CH.AIN6),
            },
            [OpCode(0x56, ADCReg.REG_ADC2MUX.value, BitMask.BYTE.value)],
        ),
        (
            {
                ADCReg.REG_INPMUX: (CH.AIN1, None),
                ADCReg.REG_ADC2MUX: (None, None),
            },
            [],
        ),
        (
            {
                ADCReg.REG_INPMUX: (CH.AIN5, None),
                ADCReg.REG_ADC2MUX: (None, CH.AIN6),
            },
            [],
        ),
        (
            {
                ADCReg.REG_INPMUX: (None, CH.AIN5),
                ADCReg.REG_ADC2MUX: (CH.AIN6, None),
            },
            [],
        ),
        (
            {
                ADCReg.REG_INPMUX: (CH.AIN1, CH.AIN2),
                ADCReg.REG_ADC2MUX: (CH.AIN3, CH.AIN4),
            },
            [
                OpCode(0x12, ADCReg.REG_INPMUX.value, BitMask.BYTE.value),
                OpCode(0x34, ADCReg.REG_ADC2MUX.value, BitMask.BYTE.value),
            ],
        ),
    ],
)
def test_generate_mux_opcodes(mux_updates, expected):
    assert generate_mux_opcodes(mux_updates) == expected


@pytest.mark.parametrize(
    "channels, rtd_enabled, err_type",
    [
        ([CH.AIN1], True, does_not_raise()),
        ([CH.AIN0, CH.AIN1, CH.AIN2, CH.AIN3, CH.AINCOM, CH.FLOAT], False, does_not_raise()),
        (
            [
                CH.AIN0,
                CH.AIN1,
                CH.AIN2,
                CH.AIN3,
                CH.AIN3,
                CH.AIN4,
                CH.AIN5,
                CH.AIN6,
                CH.AIN7,
                CH.AINCOM,
                CH.FLOAT,
            ],
            True,
            pytest.raises(ChannelNotAvailableError),
        ),
        ([CH.AIN4], True, pytest.raises(ChannelNotAvailableError)),
        (
            [
                CH.AIN4,
                CH.AIN5,
                CH.AIN6,
                CH.AIN7,
            ],
            True,
            pytest.raises(ChannelNotAvailableError),
        ),
    ],
)
def test_validate_channels_allowed(channels, rtd_enabled, err_type):
    with err_type:
        validate_channels_allowed(channels, rtd_enabled)



"""unit tests for gpio_config module"""


import pytest
from edgepi.reg_helper.reg_helper import OpCode
from edgepi.gpio.gpio_constants import (
    GpioAOutputClear,
    GpioAOutputSet,
    GpioAPinDirIn,
    GpioBOutputClear,
    GpioBOutputSet,
    GpioExpanderAddress,
    GpioAPinDirOut,
    GpioBPinDirOut
)
from edgepi.gpio.gpio_configs import (
    GpioConfigs,
    I2cPinInfo,
    generate_pin_info,
    generate_expander_pin_info,
    generate_gpiochip_pin_info
)

@pytest.mark.parametrize('config, result',
                        [(GpioConfigs.DAC.value,
                          ['dac', 'i2c', 9, 'out', 'A', GpioExpanderAddress, '/dev/i2c-10']),
                         (GpioConfigs.ADC.value,
                          ['adc', 'i2c', 2, 'out', 'B', 33, '/dev/i2c-10']),
                         (GpioConfigs.RTD.value,
                          ['rtd', 'i2c', 1, 'out', 'B', 33, '/dev/i2c-10']),
                         (GpioConfigs.LED.value,
                          ['led', 'i2c', 8, 'in', 'B', 32, '/dev/i2c-10'])
                        ])
def test_gpio_module_config(config, result):
    assert config.name == result[0]
    assert config.device == result[1]
    assert config.num_pins == result[2]
    assert config.dir == result[3]
    assert config.port == result[4]
    assert config.address == result[5]
    assert config.dev_path == result[6]

@pytest.mark.parametrize('set_output, clear_output, pin_dir, address, result',
                        [(GpioAOutputSet.SET_OUTPUT_1.value,
                          GpioAOutputClear.CLEAR_OUTPUT_1.value,
                          GpioAPinDirIn.ALL_DIR_IN.value,
                          GpioExpanderAddress.EXP_ONE.value,
                          ['AO_EN1',
                           OpCode(1, 3, 0xFE),
                           OpCode(0, 3, 0xFE),
                           OpCode(0xFF, 7, 0),
                           32]),
                         (GpioAOutputSet.SET_OUTPUT_2.value,
                          GpioBOutputClear.CLEAR_OUTPUT_2.value,
                          GpioBPinDirOut.ALL_DIR_OUT.value,
                          GpioExpanderAddress.EXP_ONE.value,
                          ['AO_EN1',
                           OpCode(2, 3, 0xFD),
                           OpCode(0, 2, 0xFD),
                           OpCode(0, 6, 0),
                           32]),
                         (GpioBOutputSet.SET_OUTPUT_3.value,
                          GpioAOutputClear.CLEAR_OUTPUT_3.value,
                          GpioAPinDirIn.ALL_DIR_IN.value,
                          GpioExpanderAddress.EXP_ONE.value,
                          ['AO_EN1',
                           OpCode(4, 2, 0xFB),
                           OpCode(0, 3, 0xFB),
                           OpCode(0xFF, 7, 0),
                           32]),
                         (GpioBOutputSet.SET_OUTPUT_4.value,
                          GpioBOutputClear.CLEAR_OUTPUT_4.value,
                          GpioBPinDirOut.ALL_DIR_OUT.value,
                          GpioExpanderAddress.EXP_TWO.value,
                          ['AO_EN1',
                           OpCode(8, 2, 0xF7),
                           OpCode(0, 2, 0xF7),
                           OpCode(0, 6, 0),
                           33])
                        ])
def test_i2c_pin_info(set_output, clear_output, pin_dir, address, result):
    pin = I2cPinInfo(set_output, clear_output, pin_dir, None, address)
    assert pin.set_code == result[1]
    assert pin.clear_code == result[2]
    assert pin.dir_out_code == result[3]
    assert pin.address == result[4]
    assert pin.is_high is None
    assert pin.is_out is None



def test_generate_pin_info_dac(config = GpioConfigs.DAC.value):
    pin_list = generate_pin_info(config)
    for pin, output_set, output_clear, output_dir in\
    zip(pin_list.items(), GpioAOutputSet, GpioAOutputClear, GpioAPinDirOut):

        if pin[0] != 'DAC_GAIN':
            assert pin[1].set_code == output_set.value
            assert pin[1].clear_code == output_clear.value
            assert pin[1].dir_out_code == output_dir.value
            assert pin[1].address == GpioExpanderAddress.EXP_ONE.value
    assert pin_list['DAC_GAIN'].set_code == GpioAOutputSet.SET_OUTPUT_1.value
    assert pin_list['DAC_GAIN'].clear_code == GpioAOutputClear.CLEAR_OUTPUT_1.value
    assert pin_list['DAC_GAIN'].dir_out_code == GpioAPinDirOut.PIN1_DIR_OUT.value
    assert pin_list['DAC_GAIN'].address == GpioExpanderAddress.EXP_TWO.value

def test_generate_pin_info_led(config = GpioConfigs.LED.value):
    pin_list = generate_pin_info(config)
    for pin, output_set, output_clear, output_dir in\
    zip(pin_list.items(), GpioBOutputSet, GpioBOutputClear, GpioBPinDirOut):

        assert pin[1].set_code == output_set.value
        assert pin[1].clear_code == output_clear.value
        assert pin[1].dir_out_code == output_dir.value
        assert pin[1].address == GpioExpanderAddress.EXP_ONE.value

def test_generate_pin_info_rtd(config = GpioConfigs.RTD.value):
    pin_list = generate_pin_info(config)
    for pin, output_set, output_clear, output_dir in\
    zip(pin_list.items(), GpioBOutputSet, GpioBOutputClear, GpioBPinDirOut):

        assert pin[1].set_code == output_set.value
        assert pin[1].clear_code == output_clear.value
        assert pin[1].dir_out_code == output_dir.value
        assert pin[1].address == GpioExpanderAddress.EXP_TWO.value

def test_generate_pin_info_adc(config = GpioConfigs.ADC.value):
    pin_list = generate_pin_info(config)
    pin_keys = list(pin_list.keys())
    assert pin_list[pin_keys[0]].set_code == GpioBOutputSet.SET_OUTPUT_2.value
    assert pin_list[pin_keys[0]].clear_code == GpioBOutputClear.CLEAR_OUTPUT_2.value
    assert pin_list[pin_keys[0]].dir_out_code == GpioBPinDirOut.PIN2_DIR_OUT.value
    assert pin_list[pin_keys[0]].address == GpioExpanderAddress.EXP_TWO.value

    assert pin_list[pin_keys[1]].set_code == GpioBOutputSet.SET_OUTPUT_3.value
    assert pin_list[pin_keys[1]].clear_code == GpioBOutputClear.CLEAR_OUTPUT_3.value
    assert pin_list[pin_keys[1]].dir_out_code == GpioBPinDirOut.PIN3_DIR_OUT.value
    assert pin_list[pin_keys[1]].address == GpioExpanderAddress.EXP_TWO.value

def test_generate_pin_info_din(config = GpioConfigs.DIN.value):
    pin_dict = generate_pin_info(config)
    pin_keys = list(pin_dict.keys())
    for pin in pin_keys:
        assert pin_dict[pin].dir == 'in'
        assert pin_dict[pin].bias == 'pull_down'

def test_generate_expander_pin_info():
    pin_dict = generate_expander_pin_info()
    result_dict = {}
    result_dict.update(generate_pin_info(GpioConfigs.DAC.value))
    result_dict.update(generate_pin_info(GpioConfigs.ADC.value))
    result_dict.update(generate_pin_info(GpioConfigs.RTD.value))
    result_dict.update(generate_pin_info(GpioConfigs.LED.value))
    result_dict.update(generate_pin_info(GpioConfigs.DOUT2.value))
    assert pin_dict == result_dict

def test_generate_gpiochip_pin_info():
    pin_dict = generate_gpiochip_pin_info()
    result_dict = {}
    result_dict.update(generate_pin_info(GpioConfigs.DIN.value))
    result_dict.update(generate_pin_info(GpioConfigs.DOUT1.value))
    assert pin_dict == result_dict



""" Utility module for GPIO configuration """

from enum import Enum, unique
from typing import Union
from dataclasses import dataclass
from edgepi.gpio.gpio_constants import (
    GpioAOutputClear,
    GpioAOutputSet,
    GpioAPinDirIn,
    GpioAPinDirOut,
    GpioBOutputClear,
    GpioBOutputSet,
    GpioBPinDirIn,
    GpioBPinDirOut,
    GpioExpanderAddress,

)


@dataclass(frozen=True)
class GpioExpanderConfig:
    """Represents peripheral information for GPIOs attached on the expander

    Attributes:
        name (str): name of config

        device (str): peripheral device

        num_pins (int): number of pins used for this configuration

        address (GpioExpanderAdreess): addresses of i2c device

        dev_path (str): device path to the peripheral device file
    """

    name: str = None
    device: str = None
    num_pins: int = None
    dir: str = None
    port: str = None
    address: int = None
    dev_path: str = None

@dataclass(frozen=True)
class GpioChipConfig:
    """
    Represents peripheral information for GPIOs
    Attributes:
        name (str): name of config

        device (str): peripheral device

        num_pins (int): number of pins used for this configuration

        dev_path (str): device path to the peripheral device file
    """
    name:str = None
    device: str = None
    num_pins: int = None
    dev_path: str = None

@unique
class GpioConfigs(Enum):
    '''Enum Class for GPIO configurations dataclasses'''
    DAC = GpioExpanderConfig(name = 'dac',
                             device='i2c',
                             num_pins=9,
                             dir='out',
                             port='A',
                             address=GpioExpanderAddress,
                             dev_path='/dev/i2c-10')
    ADC = GpioExpanderConfig(name = 'adc',
                             device='i2c',
                             num_pins=2,
                             dir='out',
                             port='B',
                             address=GpioExpanderAddress.EXP_TWO.value,
                             dev_path='/dev/i2c-10')
    RTD = GpioExpanderConfig(name = 'rtd',
                             device='i2c',
                             num_pins=1,
                             dir='out',
                             port='B',
                             address=GpioExpanderAddress.EXP_TWO.value,
                             dev_path='/dev/i2c-10')
    LED = GpioExpanderConfig(name = 'led',
                             device='i2c',
                             num_pins=8,
                             dir='in',
                             port='B',
                             address=GpioExpanderAddress.EXP_ONE.value,
                             dev_path='/dev/i2c-10')
    DIN = GpioChipConfig(name = 'din',
                         device = 'gpiochip0',
                         num_pins = 8,
                         dev_path = "/dev/gpiochip0")
    DOUT1 = GpioChipConfig(name = 'dout1',
                          device = 'gpiochip0',
                          num_pins = 2,
                          dev_path = "/dev/gpiochip0")
    DOUT2 = GpioExpanderConfig(name = 'dout2',
                               device='i2c',
                               num_pins=6,
                               dir='out',
                               port='A',
                               address=GpioExpanderAddress.EXP_TWO.value,
                               dev_path='/dev/i2c-10')

@dataclass
class I2cPinInfo:
    '''
    Represents I2C pin information

        Attributes:

            set_code (GpioAOutputSet, GpioBOutputSet): Output set code

            clear_code (GpioAOutputClear, GpioBOutputClear): Output clear code

            address (int): address of i2c device
    '''
    set_code: Union[GpioAOutputSet, GpioBOutputSet] = None
    clear_code: Union[GpioAOutputClear, GpioBOutputClear] = None
    dir_out_code: Union[GpioAPinDirOut, GpioBPinDirOut] = None
    dir_in_code: Union[GpioAPinDirIn, GpioBPinDirIn] = None
    address: int = None
    is_high: bool = None
    is_out: bool = None

@dataclass
class GpioChipPinInfo:
    """
    Represents Gpiochip Pin information
    Attributes:
        dir (str): direction of the pin, high, low, in or out
        bias (str): pull_up or pull_down
    """
    dir: str = None
    bias: str = None

class DACPins(Enum):
    "DAC GPIO Pin Names"
    AO_EN1 = 'AO_EN1'
    AO_EN2 = 'AO_EN2'
    AO_EN3 = 'AO_EN3'
    AO_EN4 = 'AO_EN4'
    AO_EN5 = 'AO_EN5'
    AO_EN6 = 'AO_EN6'
    AO_EN7 = 'AO_EN7'
    AO_EN8 = 'AO_EN8'
    DAC_GAIN = 'DAC_GAIN'


_list_of_DAC_gpios = [
    DACPins.AO_EN8.value, DACPins.AO_EN7.value, DACPins.AO_EN6.value,
    DACPins.AO_EN5.value, DACPins.AO_EN2.value, DACPins.AO_EN3.value,
    DACPins.AO_EN4.value, DACPins.AO_EN1.value, DACPins.DAC_GAIN.value,
    ]


class ADCPins(Enum):
    "ADC GPIO Pin Names"
    GNDSW_IN1 = 'GNDSW_IN1'
    GNDSW_IN2 = 'GNDSW_IN2'


_list_of_ADC_gpios = [ADCPins.GNDSW_IN1.value, ADCPins.GNDSW_IN2.value]


class RTDPins(Enum):
    "RTD GPIO Pin Names"
    RTD_EN = 'RTD_EN'


_list_of_RTD_gpios = [RTDPins.RTD_EN.value]


class LEDPins(Enum):
    "LED GPIO Pin Names"
    LED_OVR1 = 'LED_OVR1'
    LED_OVR2 = 'LED_OVR2'
    LED_OVR3 = 'LED_OVR3'
    LED_OVR4 = 'LED_OVR4'
    LED_OVR5 = 'LED_OVR5'
    LED_OVR6 = 'LED_OVR6'
    LED_OVR7 = 'LED_OVR7'
    LED_OVR8 = 'LED_OVR8'


_list_of_LED_gpios = [
    LEDPins.LED_OVR1.value, LEDPins.LED_OVR2.value,
    LEDPins.LED_OVR3.value, LEDPins.LED_OVR4.value,
    LEDPins.LED_OVR5.value, LEDPins.LED_OVR6.value,
    LEDPins.LED_OVR7.value, LEDPins.LED_OVR8.value,
]


class DINPins(Enum):
    "DIN GPIO Pin Names"
    DIN1 = 'DIN1'
    DIN2 = 'DIN2'
    DIN3 = 'DIN3'
    DIN4 = 'DIN4'
    DIN5 = 'DIN5'
    DIN6 = 'DIN6'
    DIN7 = 'DIN7'
    DIN8 = 'DIN8'


_list_of_DIN_gpios =  [
    DINPins.DIN1.value, DINPins.DIN2.value,
    DINPins.DIN3.value, DINPins.DIN4.value,
    DINPins.DIN5.value, DINPins.DIN6.value,
    DINPins.DIN7.value, DINPins.DIN8.value,
]


class DOUTPins(Enum):
    "DOUT GPIO Pin Names"
    DOUT1 = 'DOUT1'
    DOUT2 = 'DOUT2'
    DOUT3 = 'DOUT3'
    DOUT4 = 'DOUT4'
    DOUT5 = 'DOUT5'
    DOUT6 = 'DOUT6'
    DOUT7 = 'DOUT7'
    DOUT8 = 'DOUT8'


_list_of_DOUT_cpu_gpios =  [
    DOUTPins.DOUT1.value, DOUTPins.DOUT2.value
]

_list_of_DOUT_expander_gpios =[
    DOUTPins.DOUT8.value, DOUTPins.DOUT7.value,
    DOUTPins.DOUT6.value, DOUTPins.DOUT5.value,
    DOUTPins.DOUT4.value, DOUTPins.DOUT3.value]


def _generate_DAC_pins(): #pylint: disable=C0103
    ''' Generates a list I2cPinInfo dataclasses for DAC pins
        Args:
            N/A
        Returns:
            a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    '''
    pin_dict = {}
    for pin, set_code, clear_code, dir_out_code, dir_in_code in \
        zip(_list_of_DAC_gpios, GpioAOutputSet, GpioAOutputClear, GpioAPinDirOut, GpioAPinDirIn):

        pin_dict.update({pin : I2cPinInfo(set_code.value,
                                          clear_code.value,
                                          dir_out_code.value,
                                          dir_in_code.value,
                                          GpioExpanderAddress.EXP_ONE.value)
                        })
    pin_dict['DAC_GAIN'] =\
        I2cPinInfo(GpioAOutputSet.SET_OUTPUT_1.value,
                   GpioAOutputClear.CLEAR_OUTPUT_1.value,
                   GpioAPinDirOut.PIN1_DIR_OUT.value,
                   GpioAPinDirIn.PIN1_DIR_IN.value,
                   GpioExpanderAddress.EXP_TWO.value)
    return pin_dict

def _generate_LED_pins(): #pylint: disable=C0103
    '''
        Args:
            N/A
        Returns:
            a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    '''
    pin_dict = {}
    for pin, set_code, clear_code, dir_out_code, dir_in_code in\
        zip(_list_of_LED_gpios, GpioBOutputSet, GpioBOutputClear, GpioBPinDirOut, GpioBPinDirIn):

        pin_dict.update({pin : I2cPinInfo(set_code.value,
                                          clear_code.value,
                                          dir_out_code.value,
                                          dir_in_code.value,
                                          GpioExpanderAddress.EXP_ONE.value)})
    return pin_dict

def _generate_ADC_pins(): #pylint: disable=C0103
    '''
        Args:
            N/A
        Returns:
            a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    '''
    pin_dict = {}
    pin_dict.update({_list_of_ADC_gpios[0] : I2cPinInfo(GpioBOutputSet.SET_OUTPUT_2.value,
                                                        GpioBOutputClear.CLEAR_OUTPUT_2.value,
                                                        GpioBPinDirOut.PIN2_DIR_OUT.value,
                                                        GpioBPinDirIn.PIN2_DIR_IN.value,
                                                        GpioExpanderAddress.EXP_TWO.value)})
    pin_dict.update({_list_of_ADC_gpios[1] : I2cPinInfo(GpioBOutputSet.SET_OUTPUT_3.value,
                                                        GpioBOutputClear.CLEAR_OUTPUT_3.value,
                                                        GpioBPinDirOut.PIN3_DIR_OUT.value,
                                                        GpioBPinDirIn.PIN3_DIR_IN.value,
                                                        GpioExpanderAddress.EXP_TWO.value)})
    return pin_dict

def _generate_RTD_pins(): #pylint: disable=C0103
    '''
        Args:
            N/A
        Returns:
        a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    '''
    pin_dict = {}
    pin_dict.update({_list_of_RTD_gpios[0] : I2cPinInfo(GpioBOutputSet.SET_OUTPUT_1.value,
                                                        GpioBOutputClear.CLEAR_OUTPUT_1.value,
                                                        GpioBPinDirOut.PIN1_DIR_OUT.value,
                                                        GpioBPinDirIn.PIN1_DIR_IN.value,
                                                        GpioExpanderAddress.EXP_TWO.value)})
    return pin_dict

def _generate_DIN_pins(): #pylint: disable=C0103
    """
        Args:
            N/A
        Returns:
        a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    """
    pin_dict = {}
    for pin in _list_of_DIN_gpios:
        pin_dict[pin] = GpioChipPinInfo(dir = "in", bias = "pull_down")
    return pin_dict

def _generate_DOUT_cpu_pins(): #pylint: disable=C0103
    """
        Args:
            N/A
        Returns:
        a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    """
    pin_dict = {}
    for pin in _list_of_DOUT_cpu_gpios:
        pin_dict[pin] = GpioChipPinInfo(dir = "out", bias = "pull_down")
    return pin_dict

def _generate_DOUT_expander_pins(): #pylint: disable=C0103
    """
        Args:
            N/A
        Returns:
        a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    """
    pin_dict = {}
    for pin, set_code, clear_code, dir_out_code, dir_in_code in \
    zip(_list_of_DOUT_expander_gpios,GpioAOutputSet,GpioAOutputClear,GpioAPinDirOut,GpioAPinDirIn):

        pin_dict.update({pin : I2cPinInfo(set_code.value,
                                          clear_code.value,
                                          dir_out_code.value,
                                          dir_in_code.value,
                                          GpioExpanderAddress.EXP_TWO.value)})
    return pin_dict


# This function is used inside unit testing
def generate_pin_info(config: Union[GpioExpanderConfig, GpioChipConfig] = None):
    ''' Generates a dictionary of pin info dataclasses
        Args:
            config (GpioExpanderConfig): name of the module to configure the gpio pins fors
        Returns:
            a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    '''

    if config.name == GpioConfigs.DAC.value.name:
        pin_dict =  _generate_DAC_pins()
    elif config.name == GpioConfigs.LED.value.name:
        pin_dict =  _generate_LED_pins()
    elif config.name == GpioConfigs.ADC.value.name:
        pin_dict =  _generate_ADC_pins()
    elif config.name == GpioConfigs.RTD.value.name:
        pin_dict =  _generate_RTD_pins()
    elif config.name == GpioConfigs.DIN.value.name:
        pin_dict = _generate_DIN_pins()
    elif config.name == GpioConfigs.DOUT1.value.name:
        pin_dict = _generate_DOUT_cpu_pins()
    elif config.name == GpioConfigs.DOUT2.value.name:
        pin_dict = _generate_DOUT_expander_pins()
    return pin_dict

def generate_expander_pin_info():
    ''' Generates a dictionary of pin info dataclasses
        Args:
            N/A
        Returns:
            a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    '''
    pin_dict = {}
    pin_dict.update(_generate_DAC_pins())
    pin_dict.update(_generate_ADC_pins())
    pin_dict.update(_generate_RTD_pins())
    pin_dict.update(_generate_LED_pins())
    pin_dict.update(_generate_DOUT_expander_pins())
    return pin_dict

def generate_gpiochip_pin_info():
    ''' Generates a dictionary of pin info dataclasses
        Args:
            N/A
        Returns:
            a dictionary of dataclass with gpio information, {'pin_name' : pin_info_dataclass}
    '''
    pin_dict = {}
    pin_dict.update(_generate_DIN_pins())
    pin_dict.update(_generate_DOUT_cpu_pins())
    return pin_dict



""" Utility module for ADC commands """


import logging

from edgepi.adc.adc_constants import ADCComs, ADCNum


_logger = logging.getLogger(__name__)


# pylint: disable=logging-too-many-args
class ADCCommands:
    """Class representing ADC commands"""

    def __init__(self):
        _logger.info("Initializing ADC Methods")

    def read_register_command(self, address: int, num: int):
        """Trigger ADC register read"""
        self.check_for_int([address, num])
        command = [ADCComs.COM_RREG.value + address, num - 1]
        _logger.debug("RREG command to send is %s", (command + [255] * num))
        return command + [255] * num

    def write_register_command(self, address, values):
        """Trigger ADC register write"""
        self.check_for_int([address])
        all(self.check_range(value, 0, 255) for value in values)
        self.check_for_int(values)
        command = [ADCComs.COM_WREG.value + address, len(values) - 1]
        _logger.debug("WREG Command to send is %s", (command + values))
        return command + values

    def start_adc(self, adc_num: ADCNum):
        """Command to start ADC"""
        _logger.debug("Command to send is %s", ([adc_num.start_cmd]))
        return [adc_num.start_cmd]


    def stop_adc(self, adc_num: ADCNum):
        """Command to stop ADC"""
        _logger.debug("Command to send is %s", ([adc_num.stop_cmd]))
        return [adc_num.stop_cmd]


    def reset_adc(self):
        """Command to reset ADC"""
        _logger.debug("Command to send is %s", ([ADCComs.COM_RESET.value]))
        return [ADCComs.COM_RESET.value]

    @staticmethod
    def check_for_int(target_list):
        """Checks if a list contains only integer values"""
        if all(isinstance(value, int) for value in target_list) and target_list:
            return True
        raise ValueError(f"Non integer value passed {target_list}")

    @staticmethod
    def check_range(target, lower, upper):
        """Validates target is in range between a min and max value"""
        if lower <= target <= upper:
            return True
        raise ValueError(f"Target out of range {target}")
